@@ -42,6 +42,7 @@
 
 #include <qdebug.h>
 #include <qdir.h>
+#include <qscopedvaluerollback.h>
 #if defined(Q_OS_WIN)
 #include <qtimer.h>
 #endif
@@ -88,7 +89,7 @@ QT_END_NAMESPACE
 #include "qprocess_p.h"
 
 #include <qbytearray.h>
-#include <qelapsedtimer.h>
+#include <qdeadlinetimer.h>
 #include <qcoreapplication.h>
 #include <qsocketnotifier.h>
 #include <qtimer.h>
@@ -99,27 +100,12 @@ QT_END_NAMESPACE
 #include <private/qcore_unix_p.h>
 #endif
 
-#if QT_HAS_INCLUDE(<paths.h>)
+#if __has_include(<paths.h>)
 #include <paths.h>
 #endif
 
 QT_BEGIN_NAMESPACE
 
-/*!
-    \since 5.6
-
-    \macro QT_NO_PROCESS_COMBINED_ARGUMENT_START
-    \relates QProcess
-
-    Disables the
-    \l {QProcess::start(const QString &, QIODevice::OpenMode)}
-    {QProcess::start}() overload taking a single string.
-    In most cases where it is used, the user intends for the first argument
-    to be treated atomically as per the other overload.
-
-    \sa QProcess::start(const QString &command, QIODevice::OpenMode mode)
-*/
-
 /*!
     \class QProcessEnvironment
     \inmodule QtCore
@@ -147,7 +133,7 @@ QT_BEGIN_NAMESPACE
     binary data (except for the NUL character). QProcessEnvironment will preserve
     such variables, but does not support manipulating variables whose names or
     values cannot be encoded by the current locale settings (see
-    QTextCodec::codecForLocale).
+    QString::toLocal8Bit).
 
     On Windows, the variable names are case-insensitive, but case-preserving.
     QProcessEnvironment behaves accordingly.
@@ -201,6 +187,7 @@ void QProcessEnvironmentPrivate::insert(const QProcessEnvironmentPrivate &other)
         vars.insert(it.key(), it.value());
 
 #ifdef Q_OS_UNIX
+    const OrderedNameMapMutexLocker locker(this, &other);
     auto nit = other.nameMap.constBegin();
     const auto nend = other.nameMap.constEnd();
     for ( ; nit != nend; ++nit)
@@ -214,7 +201,7 @@ void QProcessEnvironmentPrivate::insert(const QProcessEnvironmentPrivate &other)
     environment variables to be removed.
 */
 QProcessEnvironment::QProcessEnvironment()
-    : d(0)
+    : d(nullptr)
 {
 }
 
@@ -274,7 +261,6 @@ bool QProcessEnvironment::operator==(const QProcessEnvironment &other) const
         return true;
     if (d) {
         if (other.d) {
-            QProcessEnvironmentPrivate::OrderedMutexLocker locker(d, other.d);
             return d->vars == other.d->vars;
         } else {
             return isEmpty();
@@ -304,7 +290,7 @@ bool QProcessEnvironment::isEmpty() const
 */
 void QProcessEnvironment::clear()
 {
-    if (d)
+    if (d.constData())
         d->vars.clear();
     // Unix: Don't clear d->nameMap, as the environment is likely to be
     // re-populated with the same keys again.
@@ -321,7 +307,6 @@ bool QProcessEnvironment::contains(const QString &name) const
 {
     if (!d)
         return false;
-    QProcessEnvironmentPrivate::MutexLocker locker(d);
     return d->vars.contains(d->prepareName(name));
 }
 
@@ -354,9 +339,9 @@ void QProcessEnvironment::insert(const QString &name, const QString &value)
 */
 void QProcessEnvironment::remove(const QString &name)
 {
-    if (d) {
-        d.detach(); // detach before prepareName()
-        d->vars.remove(d->prepareName(name));
+    if (d.constData()) {
+        QProcessEnvironmentPrivate *p = d.data();
+        p->vars.remove(p->prepareName(name));
     }
 }
 
@@ -372,7 +357,6 @@ QString QProcessEnvironment::value(const QString &name, const QString &defaultVa
     if (!d)
         return defaultValue;
 
-    QProcessEnvironmentPrivate::MutexLocker locker(d);
     const auto it = d->vars.constFind(d->prepareName(name));
     if (it == d->vars.constEnd())
         return defaultValue;
@@ -397,7 +381,6 @@ QStringList QProcessEnvironment::toStringList() const
 {
     if (!d)
         return QStringList();
-    QProcessEnvironmentPrivate::MutexLocker locker(d);
     return d->toList();
 }
 
@@ -411,7 +394,6 @@ QStringList QProcessEnvironment::keys() const
 {
     if (!d)
         return QStringList();
-    QProcessEnvironmentPrivate::MutexLocker locker(d);
     return d->keys();
 }
 
@@ -428,7 +410,6 @@ void QProcessEnvironment::insert(const QProcessEnvironment &e)
         return;
 
     // our re-impl of detach() detaches from null
-    QProcessEnvironmentPrivate::MutexLocker locker(e.d);
     d->insert(*e.d);
 }
 
@@ -440,18 +421,18 @@ void QProcessPrivate::Channel::clear()
     case PipeSource:
         Q_ASSERT(process);
         process->stdinChannel.type = Normal;
-        process->stdinChannel.process = 0;
+        process->stdinChannel.process = nullptr;
         break;
     case PipeSink:
         Q_ASSERT(process);
         process->stdoutChannel.type = Normal;
-        process->stdoutChannel.process = 0;
+        process->stdoutChannel.process = nullptr;
         break;
     }
 
     type = Normal;
     file.clear();
-    process = 0;
+    process = nullptr;
 }
 
 /*!
@@ -508,8 +489,7 @@ void QProcessPrivate::Channel::clear()
     You can also call error() to find the type of error that occurred
     last, and state() to find the current process state.
 
-    \note QProcess is not supported on VxWorks, iOS, tvOS, watchOS,
-    or the Universal Windows Platform.
+    \note QProcess is not supported on VxWorks, iOS, tvOS, or watchOS.
 
     \section1 Communicating via Channels
 
@@ -636,7 +616,8 @@ void QProcessPrivate::Channel::clear()
     process into the standard output channel (\c stdout). The
     standard error channel (\c stderr) will not receive any data. The
     standard output and standard error data of the running process
-    are interleaved.
+    are interleaved. For detached processes, the merged output of the
+    running process is forwarded onto the main process.
 
     \value ForwardedChannels QProcess forwards the output of the
     running process onto the main process. Anything the child process
@@ -693,7 +674,7 @@ void QProcessPrivate::Channel::clear()
 
     \value FailedToStart The process failed to start. Either the
     invoked program is missing, or you may have insufficient
-    permissions to invoke the program.
+    permissions or resources to invoke the program.
 
     \value Crashed The process crashed some time after starting
     successfully.
@@ -785,13 +766,6 @@ void QProcessPrivate::Channel::clear()
     \sa QProcess::CreateProcessArgumentModifier
 */
 
-/*!
-    \fn void QProcess::error(QProcess::ProcessError error)
-    \obsolete
-
-    Use errorOccurred() instead.
-*/
-
 /*!
     \fn void QProcess::errorOccurred(QProcess::ProcessError error)
     \since 5.6
@@ -814,16 +788,6 @@ void QProcessPrivate::Channel::clear()
     \a newState argument is the state QProcess changed to.
 */
 
-#if QT_DEPRECATED_SINCE(5, 13)
-/*!
-    \fn void QProcess::finished(int exitCode)
-    \obsolete
-    \overload
-
-    Use finished(int exitCode, QProcess::ExitStatus status) instead.
-*/
-#endif
-
 /*!
     \fn void QProcess::finished(int exitCode, QProcess::ExitStatus exitStatus)
 
@@ -865,27 +829,6 @@ QProcessPrivate::QProcessPrivate()
 {
     readBufferChunkSize = QRINGBUFFER_CHUNKSIZE;
     writeBufferChunkSize = QRINGBUFFER_CHUNKSIZE;
-    processChannelMode = QProcess::SeparateChannels;
-    inputChannelMode = QProcess::ManagedInputChannel;
-    processError = QProcess::UnknownError;
-    processState = QProcess::NotRunning;
-    pid = 0;
-    sequenceNumber = 0;
-    exitCode = 0;
-    exitStatus = QProcess::NormalExit;
-    startupSocketNotifier = 0;
-    deathNotifier = 0;
-    childStartedPipe[0] = INVALID_Q_PIPE;
-    childStartedPipe[1] = INVALID_Q_PIPE;
-    forkfd = -1;
-    crashed = false;
-    dying = false;
-    emittedReadyRead = false;
-    emittedBytesWritten = false;
-#ifdef Q_OS_WIN
-    stdinWriteTrigger = 0;
-    processFinishedNotifier = 0;
-#endif // Q_OS_WIN
 }
 
 /*!
@@ -906,12 +849,6 @@ void QProcessPrivate::cleanup()
 {
     q_func()->setProcessState(QProcess::NotRunning);
 #ifdef Q_OS_WIN
-    if (pid) {
-        CloseHandle(pid->hThread);
-        CloseHandle(pid->hProcess);
-        delete pid;
-        pid = 0;
-    }
     if (stdinWriteTrigger) {
         delete stdinWriteTrigger;
         stdinWriteTrigger = 0;
@@ -920,31 +857,19 @@ void QProcessPrivate::cleanup()
         delete processFinishedNotifier;
         processFinishedNotifier = 0;
     }
-
-#endif
+    if (pid) {
+        CloseHandle(pid->hThread);
+        CloseHandle(pid->hProcess);
+        delete pid;
+        pid = nullptr;
+    }
+#else
     pid = 0;
-    sequenceNumber = 0;
-    dying = false;
+#endif
 
-    if (stdoutChannel.notifier) {
-        delete stdoutChannel.notifier;
-        stdoutChannel.notifier = 0;
-    }
-    if (stderrChannel.notifier) {
-        delete stderrChannel.notifier;
-        stderrChannel.notifier = 0;
-    }
-    if (stdinChannel.notifier) {
-        delete stdinChannel.notifier;
-        stdinChannel.notifier = 0;
-    }
-    if (startupSocketNotifier) {
-        delete startupSocketNotifier;
-        startupSocketNotifier = 0;
-    }
-    if (deathNotifier) {
-        delete deathNotifier;
-        deathNotifier = 0;
+    if (stateNotifier) {
+        delete stateNotifier;
+        stateNotifier = nullptr;
     }
     closeChannel(&stdoutChannel);
     closeChannel(&stderrChannel);
@@ -998,17 +923,89 @@ void QProcessPrivate::setErrorAndEmit(QProcess::ProcessError error, const QStrin
     Q_ASSERT(error != QProcess::UnknownError);
     setError(error, description);
     emit q->errorOccurred(processError);
-#if QT_DEPRECATED_SINCE(5, 6)
-QT_WARNING_PUSH
-QT_WARNING_DISABLE_DEPRECATED
-    emit q->error(processError);
-QT_WARNING_POP
-#endif
 }
 
 /*!
     \internal
-    Returns true if we emitted readyRead().
+*/
+bool QProcessPrivate::openChannels()
+{
+    // stdin channel.
+    if (inputChannelMode == QProcess::ForwardedInputChannel) {
+        if (stdinChannel.type != Channel::Normal)
+            qWarning("QProcess::openChannels: Inconsistent stdin channel configuration");
+    } else if (!openChannel(stdinChannel)) {
+        return false;
+    }
+
+    // stdout channel.
+    if (processChannelMode == QProcess::ForwardedChannels
+            || processChannelMode == QProcess::ForwardedOutputChannel) {
+        if (stdoutChannel.type != Channel::Normal)
+            qWarning("QProcess::openChannels: Inconsistent stdout channel configuration");
+    } else if (!openChannel(stdoutChannel)) {
+        return false;
+    }
+
+    // stderr channel.
+    if (processChannelMode == QProcess::ForwardedChannels
+            || processChannelMode == QProcess::ForwardedErrorChannel
+            || processChannelMode == QProcess::MergedChannels) {
+        if (stderrChannel.type != Channel::Normal)
+            qWarning("QProcess::openChannels: Inconsistent stderr channel configuration");
+    } else if (!openChannel(stderrChannel)) {
+        return false;
+    }
+
+    return true;
+}
+
+/*!
+    \internal
+*/
+bool QProcessPrivate::openChannelsForDetached()
+{
+    // stdin channel.
+    bool needToOpen = (stdinChannel.type == Channel::Redirect
+                       || stdinChannel.type == Channel::PipeSink);
+    if (stdinChannel.type != Channel::Normal
+            && (!needToOpen
+                || inputChannelMode == QProcess::ForwardedInputChannel)) {
+        qWarning("QProcess::openChannelsForDetached: Inconsistent stdin channel configuration");
+    }
+    if (needToOpen && !openChannel(stdinChannel))
+        return false;
+
+    // stdout channel.
+    needToOpen = (stdoutChannel.type == Channel::Redirect
+                  || stdoutChannel.type == Channel::PipeSource);
+    if (stdoutChannel.type != Channel::Normal
+            && (!needToOpen
+                || processChannelMode == QProcess::ForwardedChannels
+                || processChannelMode == QProcess::ForwardedOutputChannel)) {
+        qWarning("QProcess::openChannelsForDetached: Inconsistent stdout channel configuration");
+    }
+    if (needToOpen && !openChannel(stdoutChannel))
+        return false;
+
+    // stderr channel.
+    needToOpen = (stderrChannel.type == Channel::Redirect);
+    if (stderrChannel.type != Channel::Normal
+            && (!needToOpen
+                || processChannelMode == QProcess::ForwardedChannels
+                || processChannelMode == QProcess::ForwardedErrorChannel
+                || processChannelMode == QProcess::MergedChannels)) {
+        qWarning("QProcess::openChannelsForDetached: Inconsistent stderr channel configuration");
+    }
+    if (needToOpen && !openChannel(stderrChannel))
+        return false;
+
+    return true;
+}
+
+/*!
+    \internal
+    Returns \c true if we emitted readyRead().
 */
 bool QProcessPrivate::tryReadFromChannel(Channel *channel)
 {
@@ -1043,8 +1040,6 @@ bool QProcessPrivate::tryReadFromChannel(Channel *channel)
     }
     if (readBytes == 0) {
         // EOF
-        if (channel->notifier)
-            channel->notifier->setEnabled(false);
         closeChannel(channel);
 #if defined QPROCESS_DEBUG
         qDebug("QProcessPrivate::tryReadFromChannel(%d), 0 bytes available",
@@ -1053,8 +1048,8 @@ bool QProcessPrivate::tryReadFromChannel(Channel *channel)
         return false;
     }
 #if defined QPROCESS_DEBUG
-    qDebug("QProcessPrivate::tryReadFromChannel(%d), read %d bytes from the process' output",
-           int(channel - &stdinChannel), int(readBytes));
+    qDebug("QProcessPrivate::tryReadFromChannel(%d), read %lld bytes from the process' output",
+           int(channel - &stdinChannel), readBytes);
 #endif
 
     if (channel->closed) {
@@ -1068,9 +1063,8 @@ bool QProcessPrivate::tryReadFromChannel(Channel *channel)
     if (currentReadChannel == channelIdx) {
         didRead = true;
         if (!emittedReadyRead) {
-            emittedReadyRead = true;
+            QScopedValueRollback<bool> guard(emittedReadyRead, true);
             emit q->readyRead();
-            emittedReadyRead = false;
         }
     }
     emit q->channelReadyRead(int(channelIdx));
@@ -1123,75 +1117,63 @@ bool QProcessPrivate::_q_canWrite()
 /*!
     \internal
 */
-bool QProcessPrivate::_q_processDied()
+void QProcessPrivate::_q_processDied()
 {
-    Q_Q(QProcess);
 #if defined QPROCESS_DEBUG
     qDebug("QProcessPrivate::_q_processDied()");
 #endif
-#ifdef Q_OS_UNIX
-    if (!waitForDeadChild())
-        return false;
-#endif
+
+    // in case there is data in the pipeline and this slot by chance
+    // got called before the read notifications, call these functions
+    // so the data is made available before we announce death.
 #ifdef Q_OS_WIN
-    if (processFinishedNotifier)
-        processFinishedNotifier->setEnabled(false);
     drainOutputPipes();
-#endif
-
-    // the process may have died before it got a chance to report that it was
-    // either running or stopped, so we will call _q_startupNotification() and
-    // give it a chance to emit started() or errorOccurred(FailedToStart).
-    if (processState == QProcess::Starting) {
-        if (!_q_startupNotification())
-            return true;
-    }
-
-    if (dying) {
-        // at this point we know the process is dead. prevent
-        // reentering this slot recursively by calling waitForFinished()
-        // or opening a dialog inside slots connected to the readyRead
-        // signals emitted below.
-        return true;
-    }
-    dying = true;
-
-    // in case there is data in the pipe line and this slot by chance
-    // got called before the read notifications, call these two slots
-    // so the data is made available before the process dies.
+#else
     _q_canReadStandardOutput();
     _q_canReadStandardError();
+#endif
 
+    // Slots connected to signals emitted by the functions called above
+    // might call waitFor*(), which would synchronously reap the process.
+    // So check the state to avoid trying to reap a second time.
+    if (processState != QProcess::NotRunning)
+        processFinished();
+}
+
+/*!
+    \internal
+*/
+void QProcessPrivate::processFinished()
+{
+    Q_Q(QProcess);
+#if defined QPROCESS_DEBUG
+    qDebug("QProcessPrivate::processFinished()");
+#endif
+
+#ifdef Q_OS_UNIX
+    waitForDeadChild();
+#else
     findExitCode();
+#endif
+
+    cleanup();
 
     if (crashed) {
         exitStatus = QProcess::CrashExit;
         setErrorAndEmit(QProcess::Crashed);
     }
 
-    bool wasRunning = (processState == QProcess::Running);
+    // we received EOF now:
+    emit q->readChannelFinished();
+    // in the future:
+    //emit q->standardOutputClosed();
+    //emit q->standardErrorClosed();
 
-    cleanup();
+    emit q->finished(exitCode, exitStatus);
 
-    if (wasRunning) {
-        // we received EOF now:
-        emit q->readChannelFinished();
-        // in the future:
-        //emit q->standardOutputClosed();
-        //emit q->standardErrorClosed();
-
-#if QT_DEPRECATED_SINCE(5, 13)
-QT_WARNING_PUSH
-QT_WARNING_DISABLE_DEPRECATED
-        emit q->finished(exitCode);
-QT_WARNING_POP
-#endif
-        emit q->finished(exitCode, exitStatus);
-    }
 #if defined QPROCESS_DEBUG
-    qDebug("QProcessPrivate::_q_processDied() process is dead");
+    qDebug("QProcessPrivate::processFinished(): process is dead");
 #endif
-    return true;
 }
 
 /*!
@@ -1204,8 +1186,6 @@ bool QProcessPrivate::_q_startupNotification()
     qDebug("QProcessPrivate::startupNotification()");
 #endif
 
-    if (startupSocketNotifier)
-        startupSocketNotifier->setEnabled(false);
     QString errorMessage;
     if (processStarted(&errorMessage)) {
         q->setProcessState(QProcess::Running);
@@ -1216,9 +1196,7 @@ bool QProcessPrivate::_q_startupNotification()
     q->setProcessState(QProcess::NotRunning);
     setErrorAndEmit(QProcess::FailedToStart, errorMessage);
 #ifdef Q_OS_UNIX
-    // make sure the process manager removes this entry
     waitForDeadChild();
-    findExitCode();
 #endif
     cleanup();
     return false;
@@ -1232,10 +1210,7 @@ void QProcessPrivate::closeWriteChannel()
 #if defined QPROCESS_DEBUG
     qDebug("QProcessPrivate::closeWriteChannel()");
 #endif
-    if (stdinChannel.notifier) {
-        delete stdinChannel.notifier;
-        stdinChannel.notifier = 0;
-    }
+
 #ifdef Q_OS_WIN
     // ### Find a better fix, feeding the process little by little
     // instead.
@@ -1270,39 +1245,9 @@ QProcess::~QProcess()
         kill();
         waitForFinished();
     }
-#ifdef Q_OS_UNIX
-    // make sure the process manager removes this entry
-    d->findExitCode();
-#endif
     d->cleanup();
 }
 
-#if QT_DEPRECATED_SINCE(5, 13)
-/*!
-    \obsolete
-    Returns the read channel mode of the QProcess. This function is
-    equivalent to processChannelMode()
-
-    \sa processChannelMode()
-*/
-QProcess::ProcessChannelMode QProcess::readChannelMode() const
-{
-    return processChannelMode();
-}
-
-/*!
-    \obsolete
-
-    Use setProcessChannelMode(\a mode) instead.
-
-    \sa setProcessChannelMode()
-*/
-void QProcess::setReadChannelMode(ProcessChannelMode mode)
-{
-    setProcessChannelMode(mode);
-}
-#endif
-
 /*!
     \since 4.2
 
@@ -1606,7 +1551,7 @@ QProcess::CreateProcessArgumentModifier QProcess::createProcessArgumentsModifier
     \note This function is available only on the Windows platform and requires
     C++11.
 
-    \sa QProcess::CreateProcessArgumentModifier
+    \sa QProcess::CreateProcessArgumentModifier, setChildProcessModifier()
 */
 void QProcess::setCreateProcessArgumentsModifier(CreateProcessArgumentModifier modifier)
 {
@@ -1616,6 +1561,59 @@ void QProcess::setCreateProcessArgumentsModifier(CreateProcessArgumentModifier m
 
 #endif
 
+#if defined(Q_OS_UNIX) || defined(Q_QDOC)
+/*!
+    \since 6.0
+
+    Returns the modifier function previously set by calling
+    setChildProcessModifier().
+
+    \note This function is only available on Unix platforms.
+
+    \sa setChildProcessModifier()
+*/
+std::function<void(void)> QProcess::childProcessModifier() const
+{
+    Q_D(const QProcess);
+    return d->childProcessModifier;
+}
+
+/*!
+    \since 6.0
+
+    Sets the \a modifier function for the child process, for Unix systems
+    (including \macos; for Windows, see setCreateProcessArgumentsModifier()).
+    The function contained by the \a modifier argument will be invoked in the
+    child process after \c{fork()} is completed and QProcess has set up the
+    standard file descriptors for the child process, but before \c{execve()},
+    inside start(). The modifier is useful to change certain properties of the
+    child process, such as setting up additional file descriptors or closing
+    others, changing the nice level, disconnecting from the controlling TTY,
+    etc.
+
+    The following shows an example of setting up a child process to run without
+    privileges:
+
+    \snippet code/src_corelib_io_qprocess.cpp 4
+
+    If the modifier function needs to exit the process, remember to use
+    \c{_exit()}, not \c{exit()}.
+
+    \note In multithreaded applications, this function must be careful not to
+    call any functions that may lock mutexes that may have been in use in
+    other threads (in general, using only functions defined by POSIX as
+    "async-signal-safe" is advised). Most of the Qt API is unsafe inside this
+    callback, including qDebug(), and may lead to deadlocks.
+
+    \sa childProcessModifier()
+*/
+void QProcess::setChildProcessModifier(const std::function<void(void)> &modifier)
+{
+    Q_D(QProcess);
+    d->childProcessModifier = modifier;
+}
+#endif
+
 /*!
     If QProcess has been assigned a working directory, this function returns
     the working directory that the QProcess will enter before the program has
@@ -1647,24 +1645,6 @@ void QProcess::setWorkingDirectory(const QString &dir)
     d->workingDirectory = dir;
 }
 
-
-/*!
-    \deprecated
-    Use processId() instead.
-
-    Returns the native process identifier for the running process, if
-    available.  If no process is currently running, \c 0 is returned.
-
-    \note Unlike \l processId(), pid() returns an integer on Unix and a pointer on Windows.
-
-    \sa Q_PID, processId()
-*/
-Q_PID QProcess::pid() const // ### Qt 6 remove or rename this method to processInformation()
-{
-    Q_D(const QProcess);
-    return d->pid;
-}
-
 /*!
     \since 5.3
 
@@ -1681,17 +1661,6 @@ qint64 QProcess::processId() const
 #endif
 }
 
-/*! \reimp
-
-    This function operates on the current read channel.
-
-    \sa readChannel(), setReadChannel()
-*/
-bool QProcess::canReadLine() const
-{
-    return QIODevice::canReadLine();
-}
-
 /*!
     Closes all communication with the process and kills it. After calling this
     function, QProcess will no longer emit readyRead(), and data can no
@@ -1709,16 +1678,6 @@ void QProcess::close()
     QIODevice::close();
 }
 
-/*! \reimp
-
-   Returns \c true if the process is not running, and no more data is available
-   for reading; otherwise returns \c false.
-*/
-bool QProcess::atEnd() const
-{
-    return QIODevice::atEnd();
-}
-
 /*! \reimp
 */
 bool QProcess::isSequential() const
@@ -1726,13 +1685,6 @@ bool QProcess::isSequential() const
     return true;
 }
 
-/*! \reimp
-*/
-qint64 QProcess::bytesAvailable() const
-{
-    return QIODevice::bytesAvailable();
-}
-
 /*! \reimp
 */
 qint64 QProcess::bytesToWrite() const
@@ -1859,7 +1811,7 @@ bool QProcess::waitForStarted(int msecs)
 {
     Q_D(QProcess);
     if (d->processState == QProcess::Starting)
-        return d->waitForStarted(msecs);
+        return d->waitForStarted(QDeadlineTimer(msecs));
 
     return d->processState == QProcess::Running;
 }
@@ -1876,7 +1828,15 @@ bool QProcess::waitForReadyRead(int msecs)
         return false;
     if (d->currentReadChannel == QProcess::StandardError && d->stderrChannel.closed)
         return false;
-    return d->waitForReadyRead(msecs);
+
+    QDeadlineTimer deadline(msecs);
+    if (d->processState == QProcess::Starting) {
+        bool started = d->waitForStarted(deadline);
+        if (!started)
+            return false;
+    }
+
+    return d->waitForReadyRead(deadline);
 }
 
 /*! \reimp
@@ -1886,16 +1846,15 @@ bool QProcess::waitForBytesWritten(int msecs)
     Q_D(QProcess);
     if (d->processState == QProcess::NotRunning)
         return false;
+
+    QDeadlineTimer deadline(msecs);
     if (d->processState == QProcess::Starting) {
-        QElapsedTimer stopWatch;
-        stopWatch.start();
-        bool started = waitForStarted(msecs);
+        bool started = d->waitForStarted(deadline);
         if (!started)
             return false;
-        msecs = qt_subtract_from_timeout(msecs, stopWatch.elapsed());
     }
 
-    return d->waitForBytesWritten(msecs);
+    return d->waitForBytesWritten(deadline);
 }
 
 /*!
@@ -1922,16 +1881,15 @@ bool QProcess::waitForFinished(int msecs)
     Q_D(QProcess);
     if (d->processState == QProcess::NotRunning)
         return false;
+
+    QDeadlineTimer deadline(msecs);
     if (d->processState == QProcess::Starting) {
-        QElapsedTimer stopWatch;
-        stopWatch.start();
-        bool started = waitForStarted(msecs);
+        bool started = d->waitForStarted(deadline);
         if (!started)
             return false;
-        msecs = qt_subtract_from_timeout(msecs, stopWatch.elapsed());
     }
 
-    return d->waitForFinished(msecs);
+    return d->waitForFinished(deadline);
 }
 
 /*!
@@ -1948,25 +1906,16 @@ void QProcess::setProcessState(ProcessState state)
     emit stateChanged(state, QPrivateSignal());
 }
 
+#if QT_VERSION < QT_VERSION_CHECK(7,0,0)
 /*!
-  This function is called in the child process context just before the
-    program is executed on Unix or \macos (i.e., after \c fork(), but before
-    \c execve()). Reimplement this function to do last minute initialization
-    of the child process. Example:
-
-    \snippet code/src_corelib_io_qprocess.cpp 4
-
-    You cannot exit the process (by calling exit(), for instance) from
-    this function. If you need to stop the program before it starts
-    execution, your workaround is to emit finished() and then call
-    exit().
-
-    \warning This function is called by QProcess on Unix and \macos
-    only. On Windows and QNX, it is not called.
+    \internal
 */
-void QProcess::setupChildProcess()
+auto QProcess::setupChildProcess() -> Use_setChildProcessModifier_Instead
 {
+    Q_UNREACHABLE();
+    return {};
 }
+#endif
 
 /*! \reimp
 */
@@ -2004,7 +1953,7 @@ qint64 QProcess::writeData(const char *data, qint64 len)
     }
 #endif
 
-    d->writeBuffer.append(data, len);
+    d->write(data, len);
 #ifdef Q_OS_WIN
     if (!d->stdinWriteTrigger->isActive())
         d->stdinWriteTrigger->start();
@@ -2123,6 +2072,47 @@ void QProcess::start(OpenMode mode)
     d->start(mode);
 }
 
+/*!
+    \since 6.0
+
+    Starts the command \a command in a new process.
+    The OpenMode is set to \a mode.
+
+    \a command is a single string of text containing both the program name
+    and its arguments. The arguments are separated by one or more spaces.
+    For example:
+
+    \snippet code/src_corelib_io_qprocess.cpp 5
+
+    Arguments containing spaces must be quoted to be correctly supplied to
+    the new process. For example:
+
+    \snippet code/src_corelib_io_qprocess.cpp 6
+
+    Literal quotes in the \a command string are represented by triple quotes.
+    For example:
+
+    \snippet code/src_corelib_io_qprocess.cpp 7
+
+    After the \a command string has been split and unquoted, this function
+    behaves like start().
+
+    On operating systems where the system API for passing command line
+    arguments to a subprocess natively uses a single string (Windows), one can
+    conceive command lines which cannot be passed via QProcess's portable
+    list-based API. In these rare cases you need to use setProgram() and
+    setNativeArguments() instead of this function.
+
+    \sa splitCommand()
+    \sa start()
+ */
+void QProcess::startCommand(const QString &command, OpenMode mode)
+{
+    QStringList args = splitCommand(command);
+    const QString program = args.takeFirst();
+    start(program, args, mode);
+}
+
 /*!
     \since 5.10
 
@@ -2142,11 +2132,12 @@ void QProcess::start(OpenMode mode)
     temporarily freeze.
 
     If the function is successful then *\a pid is set to the process identifier
-    of the started process. Note that the child process may exit and the PID
-    may become invalid without notice. Furthermore, after the child process
-    exits, the same PID may be recycled and used by a completely different
-    process. User code should be careful when using this variable, especially
-    if one intends to forcibly terminate the process by operating system means.
+    of the started process; otherwise, it's set to -1. Note that the child
+    process may exit and the PID may become invalid without notice.
+    Furthermore, after the child process exits, the same PID may be recycled
+    and used by a completely different process. User code should be careful
+    when using this variable, especially if one intends to forcibly terminate
+    the process by operating system means.
 
     Only the following property setters are supported by startDetached():
     \list
@@ -2158,6 +2149,8 @@ void QProcess::start(OpenMode mode)
     \li setStandardErrorFile()
     \li setStandardInputFile()
     \li setStandardOutputFile()
+    \li setProcessChannelMode(QProcess::MergedChannels)
+    \li setStandardOutputProcess()
     \li setWorkingDirectory()
     \endlist
     All other properties of the QProcess object are ignored.
@@ -2169,7 +2162,6 @@ void QProcess::start(OpenMode mode)
     \sa start()
     \sa startDetached(const QString &program, const QStringList &arguments,
                       const QString &workingDirectory, qint64 *pid)
-    \sa startDetached(const QString &command)
 */
 bool QProcess::startDetached(qint64 *pid)
 {
@@ -2192,6 +2184,8 @@ bool QProcess::startDetached(qint64 *pid)
     This method is an alias for start(), and exists only to fully implement
     the interface defined by QIODevice.
 
+    Returns \c true if the program has been started.
+
     \sa start(), setProgram(), setArguments()
 */
 bool QProcess::open(OpenMode mode)
@@ -2249,8 +2243,16 @@ void QProcessPrivate::start(QIODevice::OpenMode mode)
     startProcess();
 }
 
+/*!
+    \since 5.15
 
-static QStringList parseCombinedArgString(const QString &program)
+    Splits the string \a command into a list of tokens, and returns
+    the list.
+
+    Tokens with spaces can be surrounded by double quotes; three
+    consecutive double quotes represent the quote character itself.
+*/
+QStringList QProcess::splitCommand(QStringView command)
 {
     QStringList args;
     QString tmp;
@@ -2260,13 +2262,13 @@ static QStringList parseCombinedArgString(const QString &program)
     // handle quoting. tokens can be surrounded by double quotes
     // "hello world". three consecutive double quotes represent
     // the quote character itself.
-    for (int i = 0; i < program.size(); ++i) {
-        if (program.at(i) == QLatin1Char('"')) {
+    for (int i = 0; i < command.size(); ++i) {
+        if (command.at(i) == QLatin1Char('"')) {
             ++quoteCount;
             if (quoteCount == 3) {
                 // third consecutive quote
                 quoteCount = 0;
-                tmp += program.at(i);
+                tmp += command.at(i);
             }
             continue;
         }
@@ -2275,13 +2277,13 @@ static QStringList parseCombinedArgString(const QString &program)
                 inQuote = !inQuote;
             quoteCount = 0;
         }
-        if (!inQuote && program.at(i).isSpace()) {
+        if (!inQuote && command.at(i).isSpace()) {
             if (!tmp.isEmpty()) {
                 args += tmp;
                 tmp.clear();
             }
         } else {
-            tmp += program.at(i);
+            tmp += command.at(i);
         }
     }
     if (!tmp.isEmpty())
@@ -2290,60 +2292,6 @@ static QStringList parseCombinedArgString(const QString &program)
     return args;
 }
 
-/*!
-    \overload
-
-    Starts the command \a command in a new process.
-    The OpenMode is set to \a mode.
-
-    \a command is a single string of text containing both the program name
-    and its arguments. The arguments are separated by one or more spaces.
-    For example:
-
-    \snippet code/src_corelib_io_qprocess.cpp 5
-
-    Arguments containing spaces must be quoted to be correctly supplied to
-    the new process. For example:
-
-    \snippet code/src_corelib_io_qprocess.cpp 6
-
-    Literal quotes in the \a command string are represented by triple quotes.
-    For example:
-
-    \snippet code/src_corelib_io_qprocess.cpp 7
-
-    After the \a command string has been split and unquoted, this function
-    behaves like the overload which takes the arguments as a string list.
-
-    You can disable this overload by defining \c
-    QT_NO_PROCESS_COMBINED_ARGUMENT_START when you compile your applications.
-    This can be useful if you want to ensure that you are not splitting arguments
-    unintentionally, for example. In virtually all cases, using the other overload
-    is the preferred method.
-
-    On operating systems where the system API for passing command line
-    arguments to a subprocess natively uses a single string (Windows), one can
-    conceive command lines which cannot be passed via QProcess's portable
-    list-based API. In these rare cases you need to use setProgram() and
-    setNativeArguments() instead of this function.
-
-*/
-#if !defined(QT_NO_PROCESS_COMBINED_ARGUMENT_START)
-void QProcess::start(const QString &command, OpenMode mode)
-{
-    QStringList args = parseCombinedArgString(command);
-    if (args.isEmpty()) {
-        Q_D(QProcess);
-        d->setErrorAndEmit(QProcess::FailedToStart, tr("No program defined"));
-        return;
-    }
-
-    const QString prog = args.takeFirst();
-
-    start(prog, args, mode);
-}
-#endif
-
 /*!
     \since 5.0
 
@@ -2495,29 +2443,6 @@ int QProcess::execute(const QString &program, const QStringList &arguments)
     return process.exitStatus() == QProcess::NormalExit ? process.exitCode() : -1;
 }
 
-/*!
-    \overload
-
-    Starts the program \a command in a new process, waits for it to finish,
-    and then returns the exit code.
-
-    Argument handling is identical to the respective start() overload.
-
-    After the \a command string has been split and unquoted, this function
-    behaves like the overload which takes the arguments as a string list.
-
-    \sa start()
-*/
-int QProcess::execute(const QString &command)
-{
-    QProcess process;
-    process.setProcessChannelMode(ForwardedChannels);
-    process.start(command);
-    if (!process.waitForFinished(-1) || process.error() == FailedToStart)
-        return -2;
-    return process.exitStatus() == QProcess::NormalExit ? process.exitCode() : -1;
-}
-
 /*!
     \overload startDetached()
 
@@ -2549,45 +2474,8 @@ bool QProcess::startDetached(const QString &program,
     return process.startDetached(pid);
 }
 
-/*!
-    \internal
-*/
-bool QProcess::startDetached(const QString &program,
-                             const QStringList &arguments)
-{
-    QProcess process;
-    process.setProgram(program);
-    process.setArguments(arguments);
-    return process.startDetached();
-}
-
-/*!
-    \overload startDetached()
-
-    Starts the command \a command in a new process, and detaches from it.
-    Returns \c true on success; otherwise returns \c false.
-
-    Argument handling is identical to the respective start() overload.
-
-    After the \a command string has been split and unquoted, this function
-    behaves like the overload which takes the arguments as a string list.
-
-    \sa start(const QString &command, QIODevice::OpenMode mode)
-*/
-bool QProcess::startDetached(const QString &command)
-{
-    QStringList args = parseCombinedArgString(command);
-    if (args.isEmpty())
-        return false;
-
-    QProcess process;
-    process.setProgram(args.takeFirst());
-    process.setArguments(args);
-    return process.startDetached();
-}
-
 QT_BEGIN_INCLUDE_NAMESPACE
-#if defined(Q_OS_MACX)
+#if defined(Q_OS_MACOS)
 # include <crt_externs.h>
 # define environ (*_NSGetEnviron())
 #elif defined(QT_PLATFORM_UIKIT)
@@ -2620,7 +2508,7 @@ QT_END_INCLUDE_NAMESPACE
 QStringList QProcess::systemEnvironment()
 {
     QStringList tmp;
-    char *entry = 0;
+    char *entry = nullptr;
     int count = 0;
     while ((entry = environ[count++]))
         tmp << QString::fromLocal8Bit(entry);
@@ -2667,17 +2555,6 @@ QString QProcess::nullDevice()
 #endif
 }
 
-/*!
-    \typedef Q_PID
-    \relates QProcess
-
-    Typedef for the identifiers used to represent processes on the underlying
-    platform. On Unix, this corresponds to \l qint64; on Windows, it
-    corresponds to \c{_PROCESS_INFORMATION*}.
-
-    \sa QProcess::pid()
-*/
-
 #endif // QT_CONFIG(process)
 
 QT_END_NAMESPACE
