@@ -47,13 +47,10 @@
 #include "qset.h"
 #include "qsocketnotifier.h"
 #include "qvarlengtharray.h"
-#include "qwineventnotifier.h"
 
 #include "qelapsedtimer.h"
 #include "qcoreapplication_p.h"
 #include <private/qthread_p.h>
-#include <private/qmutexpool_p.h>
-#include <private/qwineventnotifier_p.h>
 
 QT_BEGIN_NAMESPACE
 
@@ -82,8 +79,11 @@ extern uint qGlobalPostedEventsCount();
 enum {
     WM_QT_SOCKETNOTIFIER = WM_USER,
     WM_QT_SENDPOSTEDEVENTS = WM_USER + 1,
-    WM_QT_ACTIVATENOTIFIERS = WM_USER + 2,
-    SendPostedEventsWindowsTimerId = ~1u
+    WM_QT_ACTIVATENOTIFIERS = WM_USER + 2
+};
+
+enum {
+    SendPostedEventsTimerId = ~1u
 };
 
 class QEventDispatcherWin32Private;
@@ -95,36 +95,40 @@ class QEventDispatcherWin32Private;
 LRESULT QT_WIN_CALLBACK qt_internal_proc(HWND hwnd, UINT message, WPARAM wp, LPARAM lp);
 
 QEventDispatcherWin32Private::QEventDispatcherWin32Private()
-    : threadId(GetCurrentThreadId()), interrupt(false), closingDown(false), internalHwnd(0),
-      getMessageHook(0), serialNumber(0), lastSerialNumber(0), sendPostedEventsWindowsTimerId(0),
-      wakeUps(0), activateNotifiersPosted(false), winEventNotifierActivatedEvent(NULL)
+    : interrupt(false), internalHwnd(0),
+      sendPostedEventsTimerId(0), wakeUps(0),
+      activateNotifiersPosted(false)
 {
 }
 
 QEventDispatcherWin32Private::~QEventDispatcherWin32Private()
 {
-    if (winEventNotifierActivatedEvent)
-        CloseHandle(winEventNotifierActivatedEvent);
     if (internalHwnd)
         DestroyWindow(internalHwnd);
 }
 
-void QEventDispatcherWin32Private::activateEventNotifier(QWinEventNotifier * wen)
-{
-    QEvent event(QEvent::WinEventAct);
-    QCoreApplication::sendEvent(wen, &event);
-}
-
 // This function is called by a workerthread
 void WINAPI QT_WIN_CALLBACK qt_fast_timer_proc(uint timerId, uint /*reserved*/, DWORD_PTR user, DWORD_PTR /*reserved*/, DWORD_PTR /*reserved*/)
 {
     if (!timerId) // sanity check
         return;
-    WinTimerInfo *t = (WinTimerInfo*)user;
+    auto t = reinterpret_cast<WinTimerInfo*>(user);
     Q_ASSERT(t);
     QCoreApplication::postEvent(t->dispatcher, new QTimerEvent(t->timerId));
 }
 
+static inline UINT inputQueueMask()
+{
+    UINT result = QS_ALLEVENTS;
+    // QTBUG 28513, QTBUG-29097, QTBUG-29435: QS_TOUCH, QS_POINTER became part of
+    // QS_INPUT in Windows Kit 8. They should not be used when running on pre-Windows 8.
+#if WINVER > 0x0601
+    if (QOperatingSystemVersion::current() < QOperatingSystemVersion::Windows8)
+        result &= ~(QS_TOUCH | QS_POINTER);
+#endif //  WINVER > 0x0601
+    return result;
+}
+
 LRESULT QT_WIN_CALLBACK qt_internal_proc(HWND hwnd, UINT message, WPARAM wp, LPARAM lp)
 {
     if (message == WM_NCCREATE)
@@ -136,25 +140,26 @@ LRESULT QT_WIN_CALLBACK qt_internal_proc(HWND hwnd, UINT message, WPARAM wp, LPA
     msg.wParam = wp;
     msg.lParam = lp;
     QAbstractEventDispatcher* dispatcher = QAbstractEventDispatcher::instance();
-    long result;
+    qintptr result;
     if (!dispatcher) {
         if (message == WM_TIMER)
             KillTimer(hwnd, wp);
         return 0;
-    } else if (dispatcher->filterNativeEvent(QByteArrayLiteral("windows_dispatcher_MSG"), &msg, &result)) {
-        return result;
     }
+    if (dispatcher->filterNativeEvent(QByteArrayLiteral("windows_dispatcher_MSG"), &msg, &result))
+        return result;
 
 #ifdef GWLP_USERDATA
-    QEventDispatcherWin32 *q = (QEventDispatcherWin32 *) GetWindowLongPtr(hwnd, GWLP_USERDATA);
+    auto q = reinterpret_cast<QEventDispatcherWin32 *>(GetWindowLongPtr(hwnd, GWLP_USERDATA));
 #else
-    QEventDispatcherWin32 *q = (QEventDispatcherWin32 *) GetWindowLong(hwnd, GWL_USERDATA);
+    auto q = reinterpret_cast<QEventDispatcherWin32 *>(GetWindowLong(hwnd, GWL_USERDATA));
 #endif
-    QEventDispatcherWin32Private *d = 0;
-    if (q != 0)
+    QEventDispatcherWin32Private *d = nullptr;
+    if (q != nullptr)
         d = q->d_func();
 
-    if (message == WM_QT_SOCKETNOTIFIER) {
+    switch (message) {
+    case WM_QT_SOCKETNOTIFIER: {
         // socket notifier message
         int type = -1;
         switch (WSAGETSELECTEVENT(lp)) {
@@ -174,7 +179,7 @@ LRESULT QT_WIN_CALLBACK qt_internal_proc(HWND hwnd, UINT message, WPARAM wp, LPA
             break;
         }
         if (type >= 0) {
-            Q_ASSERT(d != 0);
+            Q_ASSERT(d != nullptr);
             QSNDict *sn_vec[4] = { &d->sn_read, &d->sn_write, &d->sn_except, &d->sn_read };
             QSNDict *dict = sn_vec[type];
 
@@ -202,8 +207,9 @@ LRESULT QT_WIN_CALLBACK qt_internal_proc(HWND hwnd, UINT message, WPARAM wp, LPA
             }
         }
         return 0;
-    } else if (message == WM_QT_ACTIVATENOTIFIERS) {
-        Q_ASSERT(d != 0);
+    }
+    case WM_QT_ACTIVATENOTIFIERS: {
+        Q_ASSERT(d != nullptr);
 
         // Postpone activation if we have unhandled socket notifier messages
         // in the queue. WM_QT_ACTIVATENOTIFIERS will be posted again as a result of
@@ -226,77 +232,40 @@ LRESULT QT_WIN_CALLBACK qt_internal_proc(HWND hwnd, UINT message, WPARAM wp, LPA
         }
         d->activateNotifiersPosted = false;
         return 0;
-    } else if (message == WM_QT_SENDPOSTEDEVENTS
-               // we also use a Windows timer to send posted events when the message queue is full
-               || (message == WM_TIMER
-                   && d->sendPostedEventsWindowsTimerId != 0
-                   && wp == (uint)d->sendPostedEventsWindowsTimerId)) {
-        const int localSerialNumber = d->serialNumber.load();
-        if (localSerialNumber != d->lastSerialNumber) {
-            d->lastSerialNumber = localSerialNumber;
-            q->sendPostedEvents();
-        }
-        return 0;
-    } else if (message == WM_TIMER) {
-        Q_ASSERT(d != 0);
-        d->sendTimerEvent(wp);
-        return 0;
     }
+    case WM_TIMER:
+        Q_ASSERT(d != nullptr);
+
+        if (wp == d->sendPostedEventsTimerId)
+            q->sendPostedEvents();
+        else
+            d->sendTimerEvent(wp);
+        return 0;
+    case WM_QT_SENDPOSTEDEVENTS:
+        Q_ASSERT(d != nullptr);
+
+        // We send posted events manually, if the window procedure was invoked
+        // by the foreign event loop (e.g. from the native modal dialog).
+        // Skip sending, if the message queue is not empty.
+        // sendPostedEventsTimer will deliver posted events later.
+        static const UINT mask = inputQueueMask();
+        if (HIWORD(GetQueueStatus(mask)) == 0)
+            q->sendPostedEvents();
+        else
+            d->startPostedEventsTimer();
+        return 0;
+    } // switch (message)
 
     return DefWindowProc(hwnd, message, wp, lp);
 }
 
-static inline UINT inputTimerMask()
+void QEventDispatcherWin32Private::startPostedEventsTimer()
 {
-    UINT result = QS_TIMER | QS_INPUT | QS_RAWINPUT;
-    // QTBUG 28513, QTBUG-29097, QTBUG-29435: QS_TOUCH, QS_POINTER became part of
-    // QS_INPUT in Windows Kit 8. They should not be used when running on pre-Windows 8.
-#if WINVER > 0x0601
-    if (QOperatingSystemVersion::current() < QOperatingSystemVersion::Windows8)
-        result &= ~(QS_TOUCH | QS_POINTER);
-#endif //  WINVER > 0x0601
-    return result;
-}
-
-LRESULT QT_WIN_CALLBACK qt_GetMessageHook(int code, WPARAM wp, LPARAM lp)
-{
-    QEventDispatcherWin32 *q = qobject_cast<QEventDispatcherWin32 *>(QAbstractEventDispatcher::instance());
-    Q_ASSERT(q != 0);
-
-    if (wp == PM_REMOVE) {
-        if (q) {
-            MSG *msg = (MSG *) lp;
-            QEventDispatcherWin32Private *d = q->d_func();
-            const int localSerialNumber = d->serialNumber.load();
-            static const UINT mask = inputTimerMask();
-            if (HIWORD(GetQueueStatus(mask)) == 0) {
-                // no more input or timer events in the message queue, we can allow posted events to be sent normally now
-                if (d->sendPostedEventsWindowsTimerId != 0) {
-                    // stop the timer to send posted events, since we now allow the WM_QT_SENDPOSTEDEVENTS message
-                    KillTimer(d->internalHwnd, d->sendPostedEventsWindowsTimerId);
-                    d->sendPostedEventsWindowsTimerId = 0;
-                }
-                (void) d->wakeUps.fetchAndStoreRelease(0);
-                if (localSerialNumber != d->lastSerialNumber
-                    // if this message IS the one that triggers sendPostedEvents(), no need to post it again
-                    && (msg->hwnd != d->internalHwnd
-                        || msg->message != WM_QT_SENDPOSTEDEVENTS)) {
-                    PostMessage(d->internalHwnd, WM_QT_SENDPOSTEDEVENTS, 0, 0);
-                }
-            } else if (d->sendPostedEventsWindowsTimerId == 0
-                       && localSerialNumber != d->lastSerialNumber) {
-                // start a special timer to continue delivering posted events while
-                // there are still input and timer messages in the message queue
-                d->sendPostedEventsWindowsTimerId = SetTimer(d->internalHwnd,
-                                                             SendPostedEventsWindowsTimerId,
-                                                             0, // we specify zero, but Windows uses USER_TIMER_MINIMUM
-                                                             NULL);
-                // we don't check the return value of SetTimer()... if creating the timer failed, there's little
-                // we can do. we just have to accept that posted events will be starved
-            }
-        }
+    if (sendPostedEventsTimerId == 0) {
+        // Start a timer to deliver posted events when the message queue is emptied.
+        sendPostedEventsTimerId = SetTimer(internalHwnd, SendPostedEventsTimerId,
+                                           USER_TIMER_MINIMUM, NULL);
     }
-    return q->d_func()->getMessageHook ? CallNextHookEx(0, code, wp, lp) : 0;
 }
 
 // Provide class name and atom for the message window used by
@@ -333,7 +302,7 @@ QWindowsMessageWindowClassContext::QWindowsMessageWindowClassContext()
     wc.lpszClassName = className;
     atom = RegisterClass(&wc);
     if (!atom) {
-        qErrnoWarning("%s RegisterClass() failed", qPrintable(qClassName));
+        qErrnoWarning("%ls RegisterClass() failed", qUtf16Printable(qClassName));
         delete [] className;
         className = 0;
     }
@@ -369,9 +338,9 @@ static HWND qt_create_internal_window(const QEventDispatcherWin32 *eventDispatch
     }
 
 #ifdef GWLP_USERDATA
-    SetWindowLongPtr(wnd, GWLP_USERDATA, (LONG_PTR)eventDispatcher);
+    SetWindowLongPtr(wnd, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(eventDispatcher));
 #else
-    SetWindowLong(wnd, GWL_USERDATA, (LONG)eventDispatcher);
+    SetWindowLong(wnd, GWL_USERDATA, reinterpret_cast<LONG>(eventDispatcher));
 #endif
 
     return wnd;
@@ -425,7 +394,7 @@ void QEventDispatcherWin32Private::unregisterTimer(WinTimerInfo *t)
     } else if (t->fastTimerId != 0) {
         timeKillEvent(t->fastTimerId);
         QCoreApplicationPrivate::removePostedTimerEvent(t->dispatcher, t->timerId);
-    } else if (internalHwnd) {
+    } else {
         KillTimer(internalHwnd, t->timerId);
     }
     t->timerId = -1;
@@ -469,206 +438,138 @@ void QEventDispatcherWin32Private::postActivateSocketNotifiers()
         activateNotifiersPosted = PostMessage(internalHwnd, WM_QT_ACTIVATENOTIFIERS, 0, 0);
 }
 
-void QEventDispatcherWin32::createInternalHwnd()
-{
-    Q_D(QEventDispatcherWin32);
-
-    if (d->internalHwnd)
-        return;
-    d->internalHwnd = qt_create_internal_window(this);
-
-    installMessageHook();
-
-    // start all normal timers
-    for (int i = 0; i < d->timerVec.count(); ++i)
-        d->registerTimer(d->timerVec.at(i));
-}
-
-void QEventDispatcherWin32::installMessageHook()
-{
-    Q_D(QEventDispatcherWin32);
-
-    if (d->getMessageHook)
-        return;
-
-    // setup GetMessage hook needed to drive our posted events
-    d->getMessageHook = SetWindowsHookEx(WH_GETMESSAGE, (HOOKPROC) qt_GetMessageHook, NULL, GetCurrentThreadId());
-    if (Q_UNLIKELY(!d->getMessageHook)) {
-        int errorCode = GetLastError();
-        qFatal("Qt: INTERNAL ERROR: failed to install GetMessage hook: %d, %s",
-               errorCode, qPrintable(qt_error_string(errorCode)));
-    }
-}
-
-void QEventDispatcherWin32::uninstallMessageHook()
-{
-    Q_D(QEventDispatcherWin32);
-
-    if (d->getMessageHook)
-        UnhookWindowsHookEx(d->getMessageHook);
-    d->getMessageHook = 0;
-}
-
 QEventDispatcherWin32::QEventDispatcherWin32(QObject *parent)
-    : QAbstractEventDispatcher(*new QEventDispatcherWin32Private, parent)
+    : QEventDispatcherWin32(*new QEventDispatcherWin32Private, parent)
 {
 }
 
 QEventDispatcherWin32::QEventDispatcherWin32(QEventDispatcherWin32Private &dd, QObject *parent)
     : QAbstractEventDispatcher(dd, parent)
-{ }
+{
+    Q_D(QEventDispatcherWin32);
+
+    d->internalHwnd = qt_create_internal_window(this);
+}
 
 QEventDispatcherWin32::~QEventDispatcherWin32()
 {
 }
 
+static bool isUserInputMessage(UINT message)
+{
+    return (message >= WM_KEYFIRST && message <= WM_KEYLAST)
+        || (message >= WM_MOUSEFIRST && message <= WM_MOUSELAST)
+        || message == WM_MOUSEWHEEL
+        || message == WM_MOUSEHWHEEL
+        || message == WM_TOUCH
+#ifndef QT_NO_GESTURES
+        || message == WM_GESTURE
+        || message == WM_GESTURENOTIFY
+#endif
+// Pointer input: Exclude WM_NCPOINTERUPDATE .. WM_POINTERROUTEDRELEASED
+        || (message >= 0x0241 && message <= 0x0253)
+        || message == WM_CLOSE;
+}
+
 bool QEventDispatcherWin32::processEvents(QEventLoop::ProcessEventsFlags flags)
 {
     Q_D(QEventDispatcherWin32);
 
-    if (!d->internalHwnd) {
-        createInternalHwnd();
-        wakeUp(); // trigger a call to sendPostedEvents()
-    }
-
-    d->interrupt = false;
+    // We don't know _when_ the interrupt occurred so we have to honor it.
+    const bool wasInterrupted = d->interrupt.fetchAndStoreRelaxed(false);
     emit awake();
 
+    // To prevent livelocks, send posted events once per iteration.
+    // QCoreApplication::sendPostedEvents() takes care about recursions.
+    sendPostedEvents();
+
+    if (wasInterrupted)
+        return false;
+
+    auto threadData = d->threadData.loadRelaxed();
     bool canWait;
     bool retVal = false;
-    bool seenWM_QT_SENDPOSTEDEVENTS = false;
-    bool needWM_QT_SENDPOSTEDEVENTS = false;
     do {
-        DWORD waitRet = 0;
-        DWORD nCount = 0;
-        HANDLE *pHandles = nullptr;
-        if (d->winEventNotifierActivatedEvent) {
-            nCount = 1;
-            pHandles = &d->winEventNotifierActivatedEvent;
-        }
         QVarLengthArray<MSG> processedTimers;
-        while (!d->interrupt) {
+        while (!d->interrupt.loadRelaxed()) {
             MSG msg;
-            bool haveMessage;
 
             if (!(flags & QEventLoop::ExcludeUserInputEvents) && !d->queuedUserInputEvents.isEmpty()) {
                 // process queued user input events
-                haveMessage = true;
                 msg = d->queuedUserInputEvents.takeFirst();
-            } else if(!(flags & QEventLoop::ExcludeSocketNotifiers) && !d->queuedSocketEvents.isEmpty()) {
+            } else if (!(flags & QEventLoop::ExcludeSocketNotifiers) && !d->queuedSocketEvents.isEmpty()) {
                 // process queued socket events
-                haveMessage = true;
                 msg = d->queuedSocketEvents.takeFirst();
-            } else {
-                haveMessage = PeekMessage(&msg, 0, 0, 0, PM_REMOVE);
-                if (haveMessage) {
-                    if ((flags & QEventLoop::ExcludeUserInputEvents)
-                        && ((msg.message >= WM_KEYFIRST
-                             && msg.message <= WM_KEYLAST)
-                            || (msg.message >= WM_MOUSEFIRST
-                                && msg.message <= WM_MOUSELAST)
-                            || msg.message == WM_MOUSEWHEEL
-                            || msg.message == WM_MOUSEHWHEEL
-                            || msg.message == WM_TOUCH
-#ifndef QT_NO_GESTURES
-                            || msg.message == WM_GESTURE
-                            || msg.message == WM_GESTURENOTIFY
-#endif
-                            || msg.message == WM_CLOSE)) {
-                        // queue user input events for later processing
-                        d->queuedUserInputEvents.append(msg);
-                        continue;
-                    }
-                    if ((flags & QEventLoop::ExcludeSocketNotifiers)
-                        && (msg.message == WM_QT_SOCKETNOTIFIER && msg.hwnd == d->internalHwnd)) {
-                        // queue socket events for later processing
-                        d->queuedSocketEvents.append(msg);
-                        continue;
-                    }
-                }
-            }
-            if (!haveMessage) {
-                // no message - check for signalled objects
-                waitRet = MsgWaitForMultipleObjectsEx(nCount, pHandles, 0, QS_ALLINPUT, MWMO_ALERTABLE);
-                if ((haveMessage = (waitRet == WAIT_OBJECT_0 + nCount))) {
-                    // a new message has arrived, process it
+            } else if (PeekMessage(&msg, 0, 0, 0, PM_REMOVE)) {
+                if (flags.testFlag(QEventLoop::ExcludeUserInputEvents)
+                    && isUserInputMessage(msg.message)) {
+                    // queue user input events for later processing
+                    d->queuedUserInputEvents.append(msg);
                     continue;
                 }
-            }
-            if (haveMessage) {
-                // WinCE doesn't support hooks at all, so we have to call this by hand :(
-                if (!d->getMessageHook)
-                    (void) qt_GetMessageHook(0, PM_REMOVE, (LPARAM) &msg);
-
-                if (d->internalHwnd == msg.hwnd && msg.message == WM_QT_SENDPOSTEDEVENTS) {
-                    if (seenWM_QT_SENDPOSTEDEVENTS) {
-                        // when calling processEvents() "manually", we only want to send posted
-                        // events once
-                        needWM_QT_SENDPOSTEDEVENTS = true;
-                        continue;
-                    }
-                    seenWM_QT_SENDPOSTEDEVENTS = true;
-                } else if (msg.message == WM_TIMER) {
-                    // avoid live-lock by keeping track of the timers we've already sent
-                    bool found = false;
-                    for (int i = 0; !found && i < processedTimers.count(); ++i) {
-                        const MSG processed = processedTimers.constData()[i];
-                        found = (processed.wParam == msg.wParam && processed.hwnd == msg.hwnd && processed.lParam == msg.lParam);
-                    }
-                    if (found)
-                        continue;
-                    processedTimers.append(msg);
-                } else if (msg.message == WM_QUIT) {
-                    if (QCoreApplication::instance())
-                        QCoreApplication::instance()->quit();
-                    return false;
+                if ((flags & QEventLoop::ExcludeSocketNotifiers)
+                    && (msg.message == WM_QT_SOCKETNOTIFIER && msg.hwnd == d->internalHwnd)) {
+                    // queue socket events for later processing
+                    d->queuedSocketEvents.append(msg);
+                    continue;
                 }
-
-                if (!filterNativeEvent(QByteArrayLiteral("windows_generic_MSG"), &msg, 0)) {
-                    TranslateMessage(&msg);
-                    DispatchMessage(&msg);
-                }
-            } else if (waitRet - WAIT_OBJECT_0 < nCount) {
-                activateEventNotifiers();
+            } else if (MsgWaitForMultipleObjectsEx(0, NULL, 0, QS_ALLINPUT, MWMO_ALERTABLE)
+                       == WAIT_OBJECT_0) {
+                // a new message has arrived, process it
+                continue;
             } else {
-                // nothing todo so break
+                // nothing to do, so break
                 break;
             }
+
+            if (d->internalHwnd == msg.hwnd && msg.message == WM_QT_SENDPOSTEDEVENTS) {
+                d->startPostedEventsTimer();
+                // Set result to 'true' because the message was sent by wakeUp().
+                retVal = true;
+                continue;
+            }
+            if (msg.message == WM_TIMER) {
+                // Skip timer event intended for use inside foreign loop.
+                if (d->internalHwnd == msg.hwnd && msg.wParam == d->sendPostedEventsTimerId)
+                    continue;
+
+                // avoid live-lock by keeping track of the timers we've already sent
+                bool found = false;
+                for (int i = 0; !found && i < processedTimers.count(); ++i) {
+                    const MSG processed = processedTimers.constData()[i];
+                    found = (processed.wParam == msg.wParam && processed.hwnd == msg.hwnd && processed.lParam == msg.lParam);
+                }
+                if (found)
+                    continue;
+                processedTimers.append(msg);
+            } else if (msg.message == WM_QUIT) {
+                if (QCoreApplication::instance())
+                    QCoreApplication::instance()->quit();
+                return false;
+            }
+
+            if (!filterNativeEvent(QByteArrayLiteral("windows_generic_MSG"), &msg, 0)) {
+                TranslateMessage(&msg);
+                DispatchMessage(&msg);
+            }
             retVal = true;
         }
 
-        // still nothing - wait for message or signalled objects
+        // wait for message
         canWait = (!retVal
-                   && !d->interrupt
-                   && (flags & QEventLoop::WaitForMoreEvents));
+                   && !d->interrupt.loadRelaxed()
+                   && flags.testFlag(QEventLoop::WaitForMoreEvents)
+                   && threadData->canWaitLocked());
         if (canWait) {
             emit aboutToBlock();
-            waitRet = MsgWaitForMultipleObjectsEx(nCount, pHandles, INFINITE, QS_ALLINPUT, MWMO_ALERTABLE | MWMO_INPUTAVAILABLE);
+            MsgWaitForMultipleObjectsEx(0, NULL, INFINITE, QS_ALLINPUT, MWMO_ALERTABLE | MWMO_INPUTAVAILABLE);
             emit awake();
-            if (waitRet - WAIT_OBJECT_0 < nCount) {
-                activateEventNotifiers();
-                retVal = true;
-            }
         }
     } while (canWait);
 
-    if (!seenWM_QT_SENDPOSTEDEVENTS && (flags & QEventLoop::EventLoopExec) == 0) {
-        // when called "manually", always send posted events
-        sendPostedEvents();
-    }
-
-    if (needWM_QT_SENDPOSTEDEVENTS)
-        PostMessage(d->internalHwnd, WM_QT_SENDPOSTEDEVENTS, 0, 0);
-
     return retVal;
 }
 
-bool QEventDispatcherWin32::hasPendingEvents()
-{
-    MSG msg;
-    return qGlobalPostedEventsCount() || PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE);
-}
-
 void QEventDispatcherWin32::registerSocketNotifier(QSocketNotifier *notifier)
 {
     Q_ASSERT(notifier);
@@ -676,10 +577,11 @@ void QEventDispatcherWin32::registerSocketNotifier(QSocketNotifier *notifier)
     int type = notifier->type();
 #ifndef QT_NO_DEBUG
     if (sockfd < 0) {
-        qWarning("QSocketNotifier: Internal error");
+        qWarning("QEventDispatcherWin32::registerSocketNotifier: invalid socket identifier");
         return;
-    } else if (notifier->thread() != thread() || thread() != QThread::currentThread()) {
-        qWarning("QSocketNotifier: socket notifiers cannot be enabled from another thread");
+    }
+    if (notifier->thread() != thread() || thread() != QThread::currentThread()) {
+        qWarning("QEventDispatcherWin32: socket notifiers cannot be enabled from another thread");
         return;
     }
 #endif
@@ -698,8 +600,6 @@ void QEventDispatcherWin32::registerSocketNotifier(QSocketNotifier *notifier)
                  "same socket %d and type %s", sockfd, t[type]);
     }
 
-    createInternalHwnd();
-
     QSockNot *sn = new QSockNot;
     sn->obj = notifier;
     sn->fd  = sockfd;
@@ -736,10 +636,11 @@ void QEventDispatcherWin32::unregisterSocketNotifier(QSocketNotifier *notifier)
 #ifndef QT_NO_DEBUG
     int sockfd = notifier->socket();
     if (sockfd < 0) {
-        qWarning("QSocketNotifier: Internal error");
+        qWarning("QEventDispatcherWin32::unregisterSocketNotifier: invalid socket identifier");
         return;
-    } else if (notifier->thread() != thread() || thread() != QThread::currentThread()) {
-        qWarning("QSocketNotifier: socket notifiers cannot be disabled from another thread");
+    }
+    if (notifier->thread() != thread() || thread() != QThread::currentThread()) {
+        qWarning("QEventDispatcherWin32: socket notifiers cannot be disabled from another thread");
         return;
     }
 #endif
@@ -778,13 +679,14 @@ void QEventDispatcherWin32::doUnregisterSocketNotifier(QSocketNotifier *notifier
     delete sn;
 }
 
-void QEventDispatcherWin32::registerTimer(int timerId, int interval, Qt::TimerType timerType, QObject *object)
+void QEventDispatcherWin32::registerTimer(int timerId, qint64 interval, Qt::TimerType timerType, QObject *object)
 {
 #ifndef QT_NO_DEBUG
     if (timerId < 1 || interval < 0 || !object) {
         qWarning("QEventDispatcherWin32::registerTimer: invalid arguments");
         return;
-    } else if (object->thread() != thread() || thread() != QThread::currentThread()) {
+    }
+    if (object->thread() != thread() || thread() != QThread::currentThread()) {
         qWarning("QEventDispatcherWin32::registerTimer: timers cannot be started from another thread");
         return;
     }
@@ -806,10 +708,8 @@ void QEventDispatcherWin32::registerTimer(int timerId, int interval, Qt::TimerTy
     t->inTimerEvent = false;
     t->fastTimerId = 0;
 
-    if (d->internalHwnd)
-        d->registerTimer(t);
+    d->registerTimer(t);
 
-    d->timerVec.append(t);                      // store in timer vector
     d->timerDict.insert(t->timerId, t);          // store timers in dict
 }
 
@@ -820,23 +720,18 @@ bool QEventDispatcherWin32::unregisterTimer(int timerId)
         qWarning("QEventDispatcherWin32::unregisterTimer: invalid argument");
         return false;
     }
-    QThread *currentThread = QThread::currentThread();
-    if (thread() != currentThread) {
+    if (thread() != QThread::currentThread()) {
         qWarning("QEventDispatcherWin32::unregisterTimer: timers cannot be stopped from another thread");
         return false;
     }
 #endif
 
     Q_D(QEventDispatcherWin32);
-    if (d->timerVec.isEmpty() || timerId <= 0)
-        return false;
 
-    WinTimerInfo *t = d->timerDict.value(timerId);
+    WinTimerInfo *t = d->timerDict.take(timerId);
     if (!t)
         return false;
 
-    d->timerDict.remove(t->timerId);
-    d->timerVec.removeAll(t);
     d->unregisterTimer(t);
     return true;
 }
@@ -848,24 +743,25 @@ bool QEventDispatcherWin32::unregisterTimers(QObject *object)
         qWarning("QEventDispatcherWin32::unregisterTimers: invalid argument");
         return false;
     }
-    QThread *currentThread = QThread::currentThread();
-    if (object->thread() != thread() || thread() != currentThread) {
+    if (object->thread() != thread() || thread() != QThread::currentThread()) {
         qWarning("QEventDispatcherWin32::unregisterTimers: timers cannot be stopped from another thread");
         return false;
     }
 #endif
 
     Q_D(QEventDispatcherWin32);
-    if (d->timerVec.isEmpty())
+    if (d->timerDict.isEmpty())
         return false;
-    WinTimerInfo *t;
-    for (int i=0; i<d->timerVec.size(); i++) {
-        t = d->timerVec.at(i);
-        if (t && t->obj == object) {                // object found
-            d->timerDict.remove(t->timerId);
-            d->timerVec.removeAt(i);
+
+    auto it = d->timerDict.begin();
+    while (it != d->timerDict.end()) {
+        WinTimerInfo *t = it.value();
+        Q_ASSERT(t);
+        if (t->obj == object) {
+            it = d->timerDict.erase(it);
             d->unregisterTimer(t);
-            --i;
+        } else {
+            ++it;
         }
     }
     return true;
@@ -874,95 +770,23 @@ bool QEventDispatcherWin32::unregisterTimers(QObject *object)
 QList<QEventDispatcherWin32::TimerInfo>
 QEventDispatcherWin32::registeredTimers(QObject *object) const
 {
+#ifndef QT_NO_DEBUG
     if (!object) {
         qWarning("QEventDispatcherWin32:registeredTimers: invalid argument");
         return QList<TimerInfo>();
     }
+#endif
 
     Q_D(const QEventDispatcherWin32);
     QList<TimerInfo> list;
-    for (int i = 0; i < d->timerVec.size(); ++i) {
-        const WinTimerInfo *t = d->timerVec.at(i);
-        if (t && t->obj == object)
+    for (WinTimerInfo *t : qAsConst(d->timerDict)) {
+        Q_ASSERT(t);
+        if (t->obj == object)
             list << TimerInfo(t->timerId, t->interval, t->timerType);
     }
     return list;
 }
 
-bool QEventDispatcherWin32::registerEventNotifier(QWinEventNotifier *notifier)
-{
-    if (!notifier) {
-        qWarning("QWinEventNotifier: Internal error");
-        return false;
-    } else if (notifier->thread() != thread() || thread() != QThread::currentThread()) {
-        qWarning("QWinEventNotifier: event notifiers cannot be enabled from another thread");
-        return false;
-    }
-
-    Q_D(QEventDispatcherWin32);
-
-    if (d->winEventNotifierList.contains(notifier))
-        return true;
-
-    d->winEventNotifierList.append(notifier);
-    d->winEventNotifierListModified = true;
-
-    if (!d->winEventNotifierActivatedEvent)
-        d->winEventNotifierActivatedEvent = CreateEvent(0, TRUE, FALSE, nullptr);
-
-    return QWinEventNotifierPrivate::get(notifier)->registerWaitObject();
-}
-
-void QEventDispatcherWin32::unregisterEventNotifier(QWinEventNotifier *notifier)
-{
-    if (!notifier) {
-        qWarning("QWinEventNotifier: Internal error");
-        return;
-    } else if (notifier->thread() != thread() || thread() != QThread::currentThread()) {
-        qWarning("QWinEventNotifier: event notifiers cannot be disabled from another thread");
-        return;
-    }
-
-    Q_D(QEventDispatcherWin32);
-
-    int i = d->winEventNotifierList.indexOf(notifier);
-    if (i == -1)
-        return;
-    d->winEventNotifierList.takeAt(i);
-    d->winEventNotifierListModified = true;
-    QWinEventNotifierPrivate *nd = QWinEventNotifierPrivate::get(notifier);
-    if (nd->waitHandle)
-        nd->unregisterWaitObject();
-}
-
-void QEventDispatcherWin32::activateEventNotifiers()
-{
-    Q_D(QEventDispatcherWin32);
-    ResetEvent(d->winEventNotifierActivatedEvent);
-
-    // Activate signaled notifiers. Our winEventNotifierList can be modified in activation slots.
-    do {
-        d->winEventNotifierListModified = false;
-        for (int i = 0; i < d->winEventNotifierList.count(); ++i) {
-            QWinEventNotifier *notifier = d->winEventNotifierList.at(i);
-            QWinEventNotifierPrivate *nd = QWinEventNotifierPrivate::get(notifier);
-            if (nd->signaledCount.load() != 0) {
-                --nd->signaledCount;
-                nd->unregisterWaitObject();
-                d->activateEventNotifier(notifier);
-            }
-        }
-    } while (d->winEventNotifierListModified);
-
-    // Re-register the remaining activated notifiers.
-    for (int i = 0; i < d->winEventNotifierList.count(); ++i) {
-        QWinEventNotifier *notifier = d->winEventNotifierList.at(i);
-        QWinEventNotifierPrivate *nd = QWinEventNotifierPrivate::get(notifier);
-        if (!nd->waitHandle)
-            nd->registerWaitObject();
-    }
-}
-
 int QEventDispatcherWin32::remainingTime(int timerId)
 {
 #ifndef QT_NO_DEBUG
@@ -974,22 +798,12 @@ int QEventDispatcherWin32::remainingTime(int timerId)
 
     Q_D(QEventDispatcherWin32);
 
-    if (d->timerVec.isEmpty())
-        return -1;
-
     quint64 currentTime = qt_msectime();
 
-    WinTimerInfo *t;
-    for (int i=0; i<d->timerVec.size(); i++) {
-        t = d->timerVec.at(i);
-        if (t && t->timerId == timerId) {                // timer found
-            if (currentTime < t->timeout) {
-                // time to wait
-                return t->timeout - currentTime;
-            } else {
-                return 0;
-            }
-        }
+    WinTimerInfo *t = d->timerDict.value(timerId);
+    if (t) {
+        // timer found, return time to wait
+        return t->timeout > currentTime ? t->timeout - currentTime : 0;
     }
 
 #ifndef QT_NO_DEBUG
@@ -1002,23 +816,20 @@ int QEventDispatcherWin32::remainingTime(int timerId)
 void QEventDispatcherWin32::wakeUp()
 {
     Q_D(QEventDispatcherWin32);
-    d->serialNumber.ref();
-    if (d->internalHwnd && d->wakeUps.testAndSetAcquire(0, 1)) {
+    if (d->wakeUps.testAndSetRelaxed(0, 1)) {
         // post a WM_QT_SENDPOSTEDEVENTS to this thread if there isn't one already pending
-        PostMessage(d->internalHwnd, WM_QT_SENDPOSTEDEVENTS, 0, 0);
+        if (!PostMessage(d->internalHwnd, WM_QT_SENDPOSTEDEVENTS, 0, 0))
+            qErrnoWarning("QEventDispatcherWin32::wakeUp: Failed to post a message");
     }
 }
 
 void QEventDispatcherWin32::interrupt()
 {
     Q_D(QEventDispatcherWin32);
-    d->interrupt = true;
+    d->interrupt.storeRelaxed(true);
     wakeUp();
 }
 
-void QEventDispatcherWin32::flush()
-{ }
-
 void QEventDispatcherWin32::startingUp()
 { }
 
@@ -1036,20 +847,22 @@ void QEventDispatcherWin32::closingDown()
     Q_ASSERT(d->active_fd.isEmpty());
 
     // clean up any timers
-    for (int i = 0; i < d->timerVec.count(); ++i)
-        d->unregisterTimer(d->timerVec.at(i));
-    d->timerVec.clear();
+    for (WinTimerInfo *t : qAsConst(d->timerDict))
+        d->unregisterTimer(t);
     d->timerDict.clear();
 
     d->closingDown = true;
 
-    uninstallMessageHook();
+    if (d->sendPostedEventsTimerId != 0)
+        KillTimer(d->internalHwnd, d->sendPostedEventsTimerId);
+    d->sendPostedEventsTimerId = 0;
 }
 
 bool QEventDispatcherWin32::event(QEvent *e)
 {
     Q_D(QEventDispatcherWin32);
-    if (e->type() == QEvent::ZeroTimerEvent) {
+    switch (e->type()) {
+    case QEvent::ZeroTimerEvent: {
         QZeroTimerEvent *zte = static_cast<QZeroTimerEvent*>(e);
         WinTimerInfo *t = d->timerDict.value(zte->timerId());
         if (t) {
@@ -1071,9 +884,12 @@ bool QEventDispatcherWin32::event(QEvent *e)
             }
         }
         return true;
-    } else if (e->type() == QEvent::Timer) {
-        QTimerEvent *te = static_cast<QTimerEvent*>(e);
-        d->sendTimerEvent(te->timerId());
+    }
+    case QEvent::Timer:
+        d->sendTimerEvent(static_cast<const QTimerEvent*>(e)->timerId());
+        break;
+    default:
+        break;
     }
     return QAbstractEventDispatcher::event(e);
 }
@@ -1081,14 +897,20 @@ bool QEventDispatcherWin32::event(QEvent *e)
 void QEventDispatcherWin32::sendPostedEvents()
 {
     Q_D(QEventDispatcherWin32);
-    QCoreApplicationPrivate::sendPostedEvents(0, 0, d->threadData);
+
+    if (d->sendPostedEventsTimerId != 0)
+        KillTimer(d->internalHwnd, d->sendPostedEventsTimerId);
+    d->sendPostedEventsTimerId = 0;
+
+    // Allow posting WM_QT_SENDPOSTEDEVENTS message.
+    d->wakeUps.storeRelaxed(0);
+
+    QCoreApplicationPrivate::sendPostedEvents(0, 0, d->threadData.loadRelaxed());
 }
 
 HWND QEventDispatcherWin32::internalHwnd()
 {
-    Q_D(QEventDispatcherWin32);
-    createInternalHwnd();
-    return d->internalHwnd;
+    return d_func()->internalHwnd;
 }
 
 QT_END_NAMESPACE
