@@ -1,34 +1,37 @@
 /****************************************************************************
 **
 ** Copyright (C) 2019 The Qt Company Ltd.
-** Contact: http://www.qt.io/licensing/
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the Qt Gui module
 **
-** $QT_BEGIN_LICENSE:LGPL3$
+** $QT_BEGIN_LICENSE:LGPL$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
 ** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see http://www.qt.io/terms-conditions. For further
-** information use the contact form at http://www.qt.io/contact-us.
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
 ** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPLv3 included in the
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
 ** packaging of this file. Please review the following information to
 ** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl.html.
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
 ** GNU General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or later as published by the Free
-** Software Foundation and appearing in the file LICENSE.GPL included in
-** the packaging of this file. Please review the following information to
-** ensure the GNU General Public License version 2.0 requirements will be
-** met: http://www.gnu.org/licenses/gpl-2.0.html.
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -51,7 +54,7 @@
 #include <QtGui/qtguiglobal.h>
 #include <QSize>
 #include <QMatrix4x4>
-#include <QVector>
+#include <QList>
 #include <QVarLengthArray>
 #include <QThread>
 #include <QColor>
@@ -90,11 +93,11 @@ private:
     quint32 m_s = 0;
 };
 
-Q_DECLARE_TYPEINFO(QRhiDepthStencilClearValue, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiDepthStencilClearValue, Q_RELOCATABLE_TYPE);
 
-Q_GUI_EXPORT bool operator==(const QRhiDepthStencilClearValue &a, const QRhiDepthStencilClearValue &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT bool operator!=(const QRhiDepthStencilClearValue &a, const QRhiDepthStencilClearValue &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT uint qHash(const QRhiDepthStencilClearValue &v, uint seed = 0) Q_DECL_NOTHROW;
+Q_GUI_EXPORT bool operator==(const QRhiDepthStencilClearValue &a, const QRhiDepthStencilClearValue &b) noexcept;
+Q_GUI_EXPORT bool operator!=(const QRhiDepthStencilClearValue &a, const QRhiDepthStencilClearValue &b) noexcept;
+Q_GUI_EXPORT size_t qHash(const QRhiDepthStencilClearValue &v, size_t seed = 0) noexcept;
 #ifndef QT_NO_DEBUG_STREAM
 Q_GUI_EXPORT QDebug operator<<(QDebug, const QRhiDepthStencilClearValue &);
 #endif
@@ -122,11 +125,11 @@ private:
     float m_maxDepth = 1.0f;
 };
 
-Q_DECLARE_TYPEINFO(QRhiViewport, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiViewport, Q_RELOCATABLE_TYPE);
 
-Q_GUI_EXPORT bool operator==(const QRhiViewport &a, const QRhiViewport &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT bool operator!=(const QRhiViewport &a, const QRhiViewport &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT uint qHash(const QRhiViewport &v, uint seed = 0) Q_DECL_NOTHROW;
+Q_GUI_EXPORT bool operator==(const QRhiViewport &a, const QRhiViewport &b) noexcept;
+Q_GUI_EXPORT bool operator!=(const QRhiViewport &a, const QRhiViewport &b) noexcept;
+Q_GUI_EXPORT size_t qHash(const QRhiViewport &v, size_t seed = 0) noexcept;
 #ifndef QT_NO_DEBUG_STREAM
 Q_GUI_EXPORT QDebug operator<<(QDebug, const QRhiViewport &);
 #endif
@@ -146,11 +149,11 @@ private:
     std::array<int, 4> m_rect { { 0, 0, 0, 0 } };
 };
 
-Q_DECLARE_TYPEINFO(QRhiScissor, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiScissor, Q_RELOCATABLE_TYPE);
 
-Q_GUI_EXPORT bool operator==(const QRhiScissor &a, const QRhiScissor &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT bool operator!=(const QRhiScissor &a, const QRhiScissor &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT uint qHash(const QRhiScissor &v, uint seed = 0) Q_DECL_NOTHROW;
+Q_GUI_EXPORT bool operator==(const QRhiScissor &a, const QRhiScissor &b) noexcept;
+Q_GUI_EXPORT bool operator!=(const QRhiScissor &a, const QRhiScissor &b) noexcept;
+Q_GUI_EXPORT size_t qHash(const QRhiScissor &v, size_t seed = 0) noexcept;
 #ifndef QT_NO_DEBUG_STREAM
 Q_GUI_EXPORT QDebug operator<<(QDebug, const QRhiScissor &);
 #endif
@@ -181,11 +184,11 @@ private:
     int m_instanceStepRate = 1;
 };
 
-Q_DECLARE_TYPEINFO(QRhiVertexInputBinding, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiVertexInputBinding, Q_RELOCATABLE_TYPE);
 
-Q_GUI_EXPORT bool operator==(const QRhiVertexInputBinding &a, const QRhiVertexInputBinding &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT bool operator!=(const QRhiVertexInputBinding &a, const QRhiVertexInputBinding &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT uint qHash(const QRhiVertexInputBinding &v, uint seed = 0) Q_DECL_NOTHROW;
+Q_GUI_EXPORT bool operator==(const QRhiVertexInputBinding &a, const QRhiVertexInputBinding &b) noexcept;
+Q_GUI_EXPORT bool operator!=(const QRhiVertexInputBinding &a, const QRhiVertexInputBinding &b) noexcept;
+Q_GUI_EXPORT size_t qHash(const QRhiVertexInputBinding &v, size_t seed = 0) noexcept;
 #ifndef QT_NO_DEBUG_STREAM
 Q_GUI_EXPORT QDebug operator<<(QDebug, const QRhiVertexInputBinding &);
 #endif
@@ -200,11 +203,19 @@ public:
         Float,
         UNormByte4,
         UNormByte2,
-        UNormByte
+        UNormByte,
+        UInt4,
+        UInt3,
+        UInt2,
+        UInt,
+        SInt4,
+        SInt3,
+        SInt2,
+        SInt
     };
 
     QRhiVertexInputAttribute() = default;
-    QRhiVertexInputAttribute(int binding, int location, Format format, quint32 offset);
+    QRhiVertexInputAttribute(int binding, int location, Format format, quint32 offset, int matrixSlice = -1);
 
     int binding() const { return m_binding; }
     void setBinding(int b) { m_binding = b; }
@@ -213,23 +224,27 @@ public:
     void setLocation(int loc) { m_location = loc; }
 
     Format format() const { return m_format; }
-    void setFormt(Format f) { m_format = f; }
+    void setFormat(Format f) { m_format = f; }
 
     quint32 offset() const { return m_offset; }
     void setOffset(quint32 ofs) { m_offset = ofs; }
 
+    int matrixSlice() const { return m_matrixSlice; }
+    void setMatrixSlice(int slice) { m_matrixSlice = slice; }
+
 private:
     int m_binding = 0;
     int m_location = 0;
     Format m_format = Float4;
     quint32 m_offset = 0;
+    int m_matrixSlice = -1;
 };
 
-Q_DECLARE_TYPEINFO(QRhiVertexInputAttribute, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiVertexInputAttribute, Q_RELOCATABLE_TYPE);
 
-Q_GUI_EXPORT bool operator==(const QRhiVertexInputAttribute &a, const QRhiVertexInputAttribute &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT bool operator!=(const QRhiVertexInputAttribute &a, const QRhiVertexInputAttribute &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT uint qHash(const QRhiVertexInputAttribute &v, uint seed = 0) Q_DECL_NOTHROW;
+Q_GUI_EXPORT bool operator==(const QRhiVertexInputAttribute &a, const QRhiVertexInputAttribute &b) noexcept;
+Q_GUI_EXPORT bool operator!=(const QRhiVertexInputAttribute &a, const QRhiVertexInputAttribute &b) noexcept;
+Q_GUI_EXPORT size_t qHash(const QRhiVertexInputAttribute &v, size_t seed = 0) noexcept;
 #ifndef QT_NO_DEBUG_STREAM
 Q_GUI_EXPORT QDebug operator<<(QDebug, const QRhiVertexInputAttribute &);
 #endif
@@ -264,16 +279,16 @@ private:
     QVarLengthArray<QRhiVertexInputBinding, 8> m_bindings;
     QVarLengthArray<QRhiVertexInputAttribute, 8> m_attributes;
 
-    friend Q_GUI_EXPORT bool operator==(const QRhiVertexInputLayout &a, const QRhiVertexInputLayout &b) Q_DECL_NOTHROW;
-    friend Q_GUI_EXPORT uint qHash(const QRhiVertexInputLayout &v, uint seed) Q_DECL_NOTHROW;
+    friend Q_GUI_EXPORT bool operator==(const QRhiVertexInputLayout &a, const QRhiVertexInputLayout &b) noexcept;
+    friend Q_GUI_EXPORT size_t qHash(const QRhiVertexInputLayout &v, size_t seed) noexcept;
     friend Q_GUI_EXPORT QDebug operator<<(QDebug, const QRhiVertexInputLayout &);
 };
 
-Q_DECLARE_TYPEINFO(QRhiVertexInputLayout, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiVertexInputLayout, Q_RELOCATABLE_TYPE);
 
-Q_GUI_EXPORT bool operator==(const QRhiVertexInputLayout &a, const QRhiVertexInputLayout &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT bool operator!=(const QRhiVertexInputLayout &a, const QRhiVertexInputLayout &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT uint qHash(const QRhiVertexInputLayout &v, uint seed = 0) Q_DECL_NOTHROW;
+Q_GUI_EXPORT bool operator==(const QRhiVertexInputLayout &a, const QRhiVertexInputLayout &b) noexcept;
+Q_GUI_EXPORT bool operator!=(const QRhiVertexInputLayout &a, const QRhiVertexInputLayout &b) noexcept;
+Q_GUI_EXPORT size_t qHash(const QRhiVertexInputLayout &v, size_t seed = 0) noexcept;
 #ifndef QT_NO_DEBUG_STREAM
 Q_GUI_EXPORT QDebug operator<<(QDebug, const QRhiVertexInputLayout &);
 #endif
@@ -306,11 +321,11 @@ private:
     QShader::Variant m_shaderVariant = QShader::StandardShader;
 };
 
-Q_DECLARE_TYPEINFO(QRhiShaderStage, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiShaderStage, Q_RELOCATABLE_TYPE);
 
-Q_GUI_EXPORT bool operator==(const QRhiShaderStage &a, const QRhiShaderStage &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT bool operator!=(const QRhiShaderStage &a, const QRhiShaderStage &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT uint qHash(const QRhiShaderStage &s, uint seed = 0) Q_DECL_NOTHROW;
+Q_GUI_EXPORT bool operator==(const QRhiShaderStage &a, const QRhiShaderStage &b) noexcept;
+Q_GUI_EXPORT bool operator!=(const QRhiShaderStage &a, const QRhiShaderStage &b) noexcept;
+Q_GUI_EXPORT size_t qHash(const QRhiShaderStage &s, size_t seed = 0) noexcept;
 #ifndef QT_NO_DEBUG_STREAM
 Q_GUI_EXPORT QDebug operator<<(QDebug, const QRhiShaderStage &);
 #endif
@@ -338,7 +353,7 @@ public:
     };
     Q_DECLARE_FLAGS(StageFlags, StageFlag)
 
-    QRhiShaderResourceBinding();
+    QRhiShaderResourceBinding() = default;
 
     bool isLayoutCompatible(const QRhiShaderResourceBinding &other) const;
 
@@ -348,6 +363,12 @@ public:
 
     static QRhiShaderResourceBinding sampledTexture(int binding, StageFlags stage, QRhiTexture *tex, QRhiSampler *sampler);
 
+    struct TextureAndSampler {
+        QRhiTexture *tex;
+        QRhiSampler *sampler;
+    };
+    static QRhiShaderResourceBinding sampledTextures(int binding, StageFlags stage, int count, const TextureAndSampler *texSamplers);
+
     static QRhiShaderResourceBinding imageLoad(int binding, StageFlags stage, QRhiTexture *tex, int level);
     static QRhiShaderResourceBinding imageStore(int binding, StageFlags stage, QRhiTexture *tex, int level);
     static QRhiShaderResourceBinding imageLoadStore(int binding, StageFlags stage, QRhiTexture *tex, int level);
@@ -370,9 +391,10 @@ public:
             int maybeSize;
             bool hasDynamicOffset;
         };
+        static const int MAX_TEX_SAMPLER_ARRAY_SIZE = 16;
         struct SampledTextureData {
-            QRhiTexture *tex;
-            QRhiSampler *sampler;
+            int count;
+            TextureAndSampler texSamplers[MAX_TEX_SAMPLER_ARRAY_SIZE];
         };
         struct StorageImageData {
             QRhiTexture *tex;
@@ -400,11 +422,11 @@ private:
 
 Q_DECLARE_OPERATORS_FOR_FLAGS(QRhiShaderResourceBinding::StageFlags)
 
-Q_DECLARE_TYPEINFO(QRhiShaderResourceBinding, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiShaderResourceBinding, Q_PRIMITIVE_TYPE);
 
-Q_GUI_EXPORT bool operator==(const QRhiShaderResourceBinding &a, const QRhiShaderResourceBinding &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT bool operator!=(const QRhiShaderResourceBinding &a, const QRhiShaderResourceBinding &b) Q_DECL_NOTHROW;
-Q_GUI_EXPORT uint qHash(const QRhiShaderResourceBinding &b, uint seed = 0) Q_DECL_NOTHROW;
+Q_GUI_EXPORT bool operator==(const QRhiShaderResourceBinding &a, const QRhiShaderResourceBinding &b) noexcept;
+Q_GUI_EXPORT bool operator!=(const QRhiShaderResourceBinding &a, const QRhiShaderResourceBinding &b) noexcept;
+Q_GUI_EXPORT size_t qHash(const QRhiShaderResourceBinding &b, size_t seed = 0) noexcept;
 #ifndef QT_NO_DEBUG_STREAM
 Q_GUI_EXPORT QDebug operator<<(QDebug, const QRhiShaderResourceBinding &);
 #endif
@@ -447,7 +469,7 @@ private:
     int m_resolveLevel = 0;
 };
 
-Q_DECLARE_TYPEINFO(QRhiColorAttachment, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiColorAttachment, Q_RELOCATABLE_TYPE);
 
 class Q_GUI_EXPORT QRhiTextureRenderTargetDescription
 {
@@ -480,14 +502,15 @@ private:
     QRhiTexture *m_depthTexture = nullptr;
 };
 
-Q_DECLARE_TYPEINFO(QRhiTextureRenderTargetDescription, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiTextureRenderTargetDescription, Q_RELOCATABLE_TYPE);
 
 class Q_GUI_EXPORT QRhiTextureSubresourceUploadDescription
 {
 public:
     QRhiTextureSubresourceUploadDescription() = default;
-    QRhiTextureSubresourceUploadDescription(const QImage &image);
+    explicit QRhiTextureSubresourceUploadDescription(const QImage &image);
     QRhiTextureSubresourceUploadDescription(const void *data, int size);
+    explicit QRhiTextureSubresourceUploadDescription(const QByteArray &data);
 
     QImage image() const { return m_image; }
     void setImage(const QImage &image) { m_image = image; }
@@ -512,7 +535,7 @@ private:
     QPoint m_sourceTopLeft;
 };
 
-Q_DECLARE_TYPEINFO(QRhiTextureSubresourceUploadDescription, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiTextureSubresourceUploadDescription, Q_RELOCATABLE_TYPE);
 
 class Q_GUI_EXPORT QRhiTextureUploadEntry
 {
@@ -535,7 +558,7 @@ private:
     QRhiTextureSubresourceUploadDescription m_desc;
 };
 
-Q_DECLARE_TYPEINFO(QRhiTextureUploadEntry, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiTextureUploadEntry, Q_RELOCATABLE_TYPE);
 
 class Q_GUI_EXPORT QRhiTextureUploadDescription
 {
@@ -558,7 +581,7 @@ private:
     QVarLengthArray<QRhiTextureUploadEntry, 16> m_entries;
 };
 
-Q_DECLARE_TYPEINFO(QRhiTextureUploadDescription, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiTextureUploadDescription, Q_RELOCATABLE_TYPE);
 
 class Q_GUI_EXPORT QRhiTextureCopyDescription
 {
@@ -596,7 +619,7 @@ private:
     QPoint m_destinationTopLeft;
 };
 
-Q_DECLARE_TYPEINFO(QRhiTextureCopyDescription, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiTextureCopyDescription, Q_RELOCATABLE_TYPE);
 
 class Q_GUI_EXPORT QRhiReadbackDescription
 {
@@ -619,7 +642,7 @@ private:
     int m_level = 0;
 };
 
-Q_DECLARE_TYPEINFO(QRhiReadbackDescription, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiReadbackDescription, Q_RELOCATABLE_TYPE);
 
 struct Q_GUI_EXPORT QRhiNativeHandles
 {
@@ -647,8 +670,9 @@ public:
 
     virtual Type resourceType() const = 0;
 
-    virtual void release() = 0;
-    void releaseAndDestroyLater();
+    virtual void destroy() = 0;
+
+    void deleteLater();
 
     QByteArray name() const;
     void setName(const QByteArray &name);
@@ -681,6 +705,11 @@ public:
     };
     Q_DECLARE_FLAGS(UsageFlags, UsageFlag)
 
+    struct NativeBuffer {
+        const void *objects[3];
+        int slotCount;
+    };
+
     QRhiResource::Type resourceType() const override;
 
     Type type() const { return m_type; }
@@ -692,7 +721,12 @@ public:
     int size() const { return m_size; }
     void setSize(int sz) { m_size = sz; }
 
-    virtual bool build() = 0;
+    virtual bool create() = 0;
+
+    virtual NativeBuffer nativeBuffer();
+
+    virtual char *beginFullDynamicBufferUpdateForCurrentFrame();
+    virtual void endFullDynamicBufferUpdateForCurrentFrame();
 
 protected:
     QRhiBuffer(QRhiImplementation *rhi, Type type_, UsageFlags usage_, int size_);
@@ -713,7 +747,9 @@ public:
         sRGB = 1 << 4,
         UsedAsTransferSource = 1 << 5,
         UsedWithGenerateMips = 1 << 6,
-        UsedWithLoadStore = 1 << 7
+        UsedWithLoadStore = 1 << 7,
+        UsedAsCompressedAtlas = 1 << 8,
+        ExternalOES = 1 << 9
     };
     Q_DECLARE_FLAGS(Flags, Flag)
 
@@ -723,13 +759,18 @@ public:
         RGBA8,
         BGRA8,
         R8,
+        RG8,
         R16,
         RED_OR_ALPHA8,
 
         RGBA16F,
         RGBA32F,
+        R16F,
+        R32F,
 
         D16,
+        D24,
+        D24S8,
         D32F,
 
         BC1,
@@ -760,6 +801,11 @@ public:
         ASTC_12x12
     };
 
+    struct NativeTexture {
+        quint64 object;
+        int layout;
+    };
+
     QRhiResource::Type resourceType() const override;
 
     Format format() const { return m_format; }
@@ -774,9 +820,10 @@ public:
     int sampleCount() const { return m_sampleCount; }
     void setSampleCount(int s) { m_sampleCount = s; }
 
-    virtual bool build() = 0;
-    virtual const QRhiNativeHandles *nativeHandles();
-    virtual bool buildFrom(const QRhiNativeHandles *src);
+    virtual bool create() = 0;
+    virtual NativeTexture nativeTexture();
+    virtual bool createFrom(NativeTexture src);
+    virtual void setNativeLayout(int layout);
 
 protected:
     QRhiTexture(QRhiImplementation *rhi, Format format_, const QSize &pixelSize_,
@@ -801,9 +848,7 @@ public:
     enum AddressMode {
         Repeat,
         ClampToEdge,
-        Border,
         Mirror,
-        MirrorOnce
     };
 
     enum CompareOp {
@@ -840,12 +885,12 @@ public:
     CompareOp textureCompareOp() const { return m_compareOp; }
     void setTextureCompareOp(CompareOp op) { m_compareOp = op; }
 
-    virtual bool build() = 0;
+    virtual bool create() = 0;
 
 protected:
     QRhiSampler(QRhiImplementation *rhi,
                 Filter magFilter_, Filter minFilter_, Filter mipmapMode_,
-                AddressMode u_, AddressMode v_);
+                AddressMode u_, AddressMode v_, AddressMode w_);
     Filter m_magFilter;
     Filter m_minFilter;
     Filter m_mipmapMode;
@@ -882,17 +927,18 @@ public:
     Flags flags() const { return m_flags; }
     void setFlags(Flags h) { m_flags = h; }
 
-    virtual bool build() = 0;
+    virtual bool create() = 0;
 
     virtual QRhiTexture::Format backingFormat() const = 0;
 
 protected:
     QRhiRenderBuffer(QRhiImplementation *rhi, Type type_, const QSize &pixelSize_,
-                     int sampleCount_, Flags flags_);
+                     int sampleCount_, Flags flags_, QRhiTexture::Format backingFormatHint_);
     Type m_type;
     QSize m_pixelSize;
     int m_sampleCount;
     Flags m_flags;
+    QRhiTexture::Format m_backingFormatHint;
 };
 
 Q_DECLARE_OPERATORS_FOR_FLAGS(QRhiRenderBuffer::Flags)
@@ -902,8 +948,11 @@ class Q_GUI_EXPORT QRhiRenderPassDescriptor : public QRhiResource
 public:
     QRhiResource::Type resourceType() const override;
 
+    virtual bool isCompatible(const QRhiRenderPassDescriptor *other) const = 0;
     virtual const QRhiNativeHandles *nativeHandles();
 
+    virtual QRhiRenderPassDescriptor *newCompatibleRenderPassDescriptor() const = 0;
+
 protected:
     QRhiRenderPassDescriptor(QRhiImplementation *rhi);
 };
@@ -944,7 +993,7 @@ public:
 
     virtual QRhiRenderPassDescriptor *newCompatibleRenderPassDescriptor() = 0;
 
-    virtual bool build() = 0;
+    virtual bool create() = 0;
 
 protected:
     QRhiTextureRenderTarget(QRhiImplementation *rhi, const QRhiTextureRenderTargetDescription &desc_, Flags flags_);
@@ -973,11 +1022,14 @@ public:
 
     bool isLayoutCompatible(const QRhiShaderResourceBindings *other) const;
 
-    virtual bool build() = 0;
+    virtual bool create() = 0;
 
 protected:
     QRhiShaderResourceBindings(QRhiImplementation *rhi);
-    QVarLengthArray<QRhiShaderResourceBinding, 8> m_bindings;
+    QVarLengthArray<QRhiShaderResourceBinding, 16> m_bindings;
+    uint m_layoutDescHash = 0;
+    QVarLengthArray<uint, 16 * 3> m_layoutDesc;
+    friend class QRhiImplementation;
 #ifndef QT_NO_DEBUG_STREAM
     friend Q_GUI_EXPORT QDebug operator<<(QDebug, const QRhiShaderResourceBindings &);
 #endif
@@ -993,7 +1045,8 @@ public:
     enum Flag {
         UsesBlendConstants = 1 << 0,
         UsesStencilRef = 1 << 1,
-        UsesScissor = 1 << 2
+        UsesScissor = 1 << 2,
+        CompileShadersWithDebugInfo = 1 << 3
     };
     Q_DECLARE_FLAGS(Flags, Flag)
 
@@ -1149,6 +1202,12 @@ public:
     float lineWidth() const { return m_lineWidth; }
     void setLineWidth(float width) { m_lineWidth = width; }
 
+    int depthBias() const { return m_depthBias; }
+    void setDepthBias(int bias) { m_depthBias = bias; }
+
+    float slopeScaledDepthBias() const { return m_slopeScaledDepthBias; }
+    void setSlopeScaledDepthBias(float bias) { m_slopeScaledDepthBias = bias; }
+
     void setShaderStages(std::initializer_list<QRhiShaderStage> list) { m_shaderStages = list; }
     template<typename InputIterator>
     void setShaderStages(InputIterator first, InputIterator last)
@@ -1168,7 +1227,7 @@ public:
     QRhiRenderPassDescriptor *renderPassDescriptor() const { return m_renderPassDesc; }
     void setRenderPassDescriptor(QRhiRenderPassDescriptor *desc) { m_renderPassDesc = desc; }
 
-    virtual bool build() = 0;
+    virtual bool create() = 0;
 
 protected:
     QRhiGraphicsPipeline(QRhiImplementation *rhi);
@@ -1187,6 +1246,8 @@ protected:
     quint32 m_stencilWriteMask = 0xFF;
     int m_sampleCount = 1;
     float m_lineWidth = 1.0f;
+    int m_depthBias = 0;
+    float m_slopeScaledDepthBias = 0.0f;
     QVarLengthArray<QRhiShaderStage, 4> m_shaderStages;
     QRhiVertexInputLayout m_vertexInputLayout;
     QRhiShaderResourceBindings *m_shaderResourceBindings = nullptr;
@@ -1195,7 +1256,7 @@ protected:
 
 Q_DECLARE_OPERATORS_FOR_FLAGS(QRhiGraphicsPipeline::Flags)
 Q_DECLARE_OPERATORS_FOR_FLAGS(QRhiGraphicsPipeline::ColorMask)
-Q_DECLARE_TYPEINFO(QRhiGraphicsPipeline::TargetBlend, Q_MOVABLE_TYPE);
+Q_DECLARE_TYPEINFO(QRhiGraphicsPipeline::TargetBlend, Q_RELOCATABLE_TYPE);
 
 class Q_GUI_EXPORT QRhiSwapChain : public QRhiResource
 {
@@ -1233,7 +1294,7 @@ public:
     virtual QRhiRenderTarget *currentFrameRenderTarget() = 0;
     virtual QSize surfacePixelSize() = 0;
     virtual QRhiRenderPassDescriptor *newCompatibleRenderPassDescriptor() = 0;
-    virtual bool buildOrResize() = 0;
+    virtual bool createOrResize() = 0;
 
 protected:
     QRhiSwapChain(QRhiImplementation *rhi);
@@ -1250,8 +1311,16 @@ Q_DECLARE_OPERATORS_FOR_FLAGS(QRhiSwapChain::Flags)
 class Q_GUI_EXPORT QRhiComputePipeline : public QRhiResource
 {
 public:
+    enum Flag {
+        CompileShadersWithDebugInfo = 1 << 0
+    };
+    Q_DECLARE_FLAGS(Flags, Flag)
+
     QRhiResource::Type resourceType() const override;
-    virtual bool build() = 0;
+    virtual bool create() = 0;
+
+    Flags flags() const { return m_flags; }
+    void setFlags(Flags f) { m_flags = f; }
 
     QRhiShaderStage shaderStage() const { return m_shaderStage; }
     void setShaderStage(const QRhiShaderStage &stage) { m_shaderStage = stage; }
@@ -1261,10 +1330,13 @@ public:
 
 protected:
     QRhiComputePipeline(QRhiImplementation *rhi);
+    Flags m_flags;
     QRhiShaderStage m_shaderStage;
     QRhiShaderResourceBindings *m_shaderResourceBindings = nullptr;
 };
 
+Q_DECLARE_OPERATORS_FOR_FLAGS(QRhiComputePipeline::Flags)
+
 class Q_GUI_EXPORT QRhiCommandBuffer : public QRhiResource
 {
 public:
@@ -1273,6 +1345,12 @@ public:
         IndexUInt32
     };
 
+    enum BeginPassFlag {
+        ExternalContent = 0x01,
+        DoNotTrackResourcesForCompute = 0x02
+    };
+    Q_DECLARE_FLAGS(BeginPassFlags, BeginPassFlag)
+
     QRhiResource::Type resourceType() const override;
 
     void resourceUpdate(QRhiResourceUpdateBatch *resourceUpdates);
@@ -1280,7 +1358,8 @@ public:
     void beginPass(QRhiRenderTarget *rt,
                    const QColor &colorClearValue,
                    const QRhiDepthStencilClearValue &depthStencilClearValue,
-                   QRhiResourceUpdateBatch *resourceUpdates = nullptr);
+                   QRhiResourceUpdateBatch *resourceUpdates = nullptr,
+                   BeginPassFlags flags = {});
     void endPass(QRhiResourceUpdateBatch *resourceUpdates = nullptr);
 
     void setGraphicsPipeline(QRhiGraphicsPipeline *ps);
@@ -1313,7 +1392,7 @@ public:
     void debugMarkEnd();
     void debugMarkMsg(const QByteArray &msg);
 
-    void beginComputePass(QRhiResourceUpdateBatch *resourceUpdates = nullptr);
+    void beginComputePass(QRhiResourceUpdateBatch *resourceUpdates = nullptr, BeginPassFlags flags = {});
     void endComputePass(QRhiResourceUpdateBatch *resourceUpdates = nullptr);
     void setComputePipeline(QRhiComputePipeline *ps);
     void dispatch(int x, int y, int z);
@@ -1326,6 +1405,8 @@ protected:
     QRhiCommandBuffer(QRhiImplementation *rhi);
 };
 
+Q_DECLARE_OPERATORS_FOR_FLAGS(QRhiCommandBuffer::BeginPassFlags)
+
 struct Q_GUI_EXPORT QRhiReadbackResult
 {
     std::function<void()> completed = nullptr;
@@ -1348,6 +1429,7 @@ public:
     void release();
 
     void merge(QRhiResourceUpdateBatch *other);
+    bool hasOptimalCapacity() const;
 
     void updateDynamicBuffer(QRhiBuffer *buf, int offset, int size, const void *data);
     void uploadStaticBuffer(QRhiBuffer *buf, int offset, int size, const void *data);
@@ -1357,7 +1439,7 @@ public:
     void uploadTexture(QRhiTexture *tex, const QImage &image);
     void copyTexture(QRhiTexture *dst, QRhiTexture *src, const QRhiTextureCopyDescription &desc = QRhiTextureCopyDescription());
     void readBackTexture(const QRhiReadbackDescription &rb, QRhiReadbackResult *result);
-    void generateMips(QRhiTexture *tex, int layer = 0);
+    void generateMips(QRhiTexture *tex);
 
 private:
     QRhiResourceUpdateBatch(QRhiImplementation *rhi);
@@ -1367,6 +1449,29 @@ private:
     friend class QRhi;
 };
 
+struct Q_GUI_EXPORT QRhiDriverInfo
+{
+    enum DeviceType {
+        UnknownDevice,
+        IntegratedDevice,
+        DiscreteDevice,
+        ExternalDevice,
+        VirtualDevice,
+        CpuDevice
+    };
+
+    QByteArray deviceName;
+    quint64 deviceId = 0;
+    quint64 vendorId = 0;
+    DeviceType deviceType = UnknownDevice;
+};
+
+Q_DECLARE_TYPEINFO(QRhiDriverInfo, Q_RELOCATABLE_TYPE);
+
+#ifndef QT_NO_DEBUG_STREAM
+Q_GUI_EXPORT QDebug operator<<(QDebug, const QRhiDriverInfo &);
+#endif
+
 struct Q_GUI_EXPORT QRhiInitParams
 {
 };
@@ -1385,7 +1490,8 @@ public:
     enum Flag {
         EnableProfiling = 1 << 0,
         EnableDebugMarkers = 1 << 1,
-        PreferSoftwareRenderer = 1 << 2
+        PreferSoftwareRenderer = 1 << 2,
+        EnablePipelineCacheDataSave = 1 << 3
     };
     Q_DECLARE_FLAGS(Flags, Flag)
 
@@ -1416,11 +1522,16 @@ public:
         BaseInstance,
         TriangleFanTopology,
         ReadBackNonUniformBuffer,
-        ReadBackNonBaseMipLevel
+        ReadBackNonBaseMipLevel,
+        TexelFetch,
+        RenderToNonBaseMipLevel,
+        IntAttributes,
+        ScreenSpaceDerivatives,
+        ReadBackAnyTextureFormat,
+        PipelineCacheDataLoadSave
     };
 
     enum BeginFrameFlag {
-        ExternalContentsInPass = 0x01
     };
     Q_DECLARE_FLAGS(BeginFrameFlags, BeginFrameFlag)
 
@@ -1433,17 +1544,25 @@ public:
         TextureSizeMin = 1,
         TextureSizeMax,
         MaxColorAttachments,
-        FramesInFlight
+        FramesInFlight,
+        MaxAsyncReadbackFrames,
+        MaxThreadGroupsPerDimension,
+        MaxThreadsPerThreadGroup,
+        MaxThreadGroupX,
+        MaxThreadGroupY,
+        MaxThreadGroupZ
     };
 
     ~QRhi();
 
     static QRhi *create(Implementation impl,
                         QRhiInitParams *params,
-                        Flags flags = Flags(),
+                        Flags flags = {},
                         QRhiNativeHandles *importDevice = nullptr);
 
     Implementation backend() const;
+    const char *backendName() const;
+    QRhiDriverInfo driverInfo() const;
     QThread *thread() const;
 
     using CleanupCallback = std::function<void(QRhi *)>;
@@ -1461,34 +1580,38 @@ public:
     QRhiRenderBuffer *newRenderBuffer(QRhiRenderBuffer::Type type,
                                       const QSize &pixelSize,
                                       int sampleCount = 1,
-                                      QRhiRenderBuffer::Flags flags = QRhiRenderBuffer::Flags());
+                                      QRhiRenderBuffer::Flags flags = {},
+                                      QRhiTexture::Format backingFormatHint = QRhiTexture::UnknownFormat);
 
     QRhiTexture *newTexture(QRhiTexture::Format format,
                             const QSize &pixelSize,
                             int sampleCount = 1,
-                            QRhiTexture::Flags flags = QRhiTexture::Flags());
+                            QRhiTexture::Flags flags = {});
 
-    QRhiSampler *newSampler(QRhiSampler::Filter magFilter, QRhiSampler::Filter minFilter,
+    QRhiSampler *newSampler(QRhiSampler::Filter magFilter,
+                            QRhiSampler::Filter minFilter,
                             QRhiSampler::Filter mipmapMode,
-                            QRhiSampler::AddressMode u, QRhiSampler::AddressMode v);
+                            QRhiSampler::AddressMode addressU,
+                            QRhiSampler::AddressMode addressV,
+                            QRhiSampler::AddressMode addressW = QRhiSampler::Repeat);
 
     QRhiTextureRenderTarget *newTextureRenderTarget(const QRhiTextureRenderTargetDescription &desc,
-                                                    QRhiTextureRenderTarget::Flags flags = QRhiTextureRenderTarget::Flags());
+                                                    QRhiTextureRenderTarget::Flags flags = {});
 
     QRhiSwapChain *newSwapChain();
-    FrameOpResult beginFrame(QRhiSwapChain *swapChain, BeginFrameFlags flags = BeginFrameFlags());
-    FrameOpResult endFrame(QRhiSwapChain *swapChain, EndFrameFlags flags = EndFrameFlags());
+    FrameOpResult beginFrame(QRhiSwapChain *swapChain, BeginFrameFlags flags = {});
+    FrameOpResult endFrame(QRhiSwapChain *swapChain, EndFrameFlags flags = {});
     bool isRecordingFrame() const;
     int currentFrameSlot() const;
 
-    FrameOpResult beginOffscreenFrame(QRhiCommandBuffer **cb, BeginFrameFlags flags = BeginFrameFlags());
-    FrameOpResult endOffscreenFrame(EndFrameFlags flags = EndFrameFlags());
+    FrameOpResult beginOffscreenFrame(QRhiCommandBuffer **cb, BeginFrameFlags flags = {});
+    FrameOpResult endOffscreenFrame(EndFrameFlags flags = {});
 
     QRhi::FrameOpResult finish();
 
     QRhiResourceUpdateBatch *nextResourceUpdateBatch();
 
-    QVector<int> supportedSampleCounts() const;
+    QList<int> supportedSampleCounts() const;
 
     int ubufAlignment() const;
     int ubufAligned(int v) const;
@@ -1502,7 +1625,7 @@ public:
 
     QMatrix4x4 clipSpaceCorrMatrix() const;
 
-    bool isTextureFormatSupported(QRhiTexture::Format format, QRhiTexture::Flags flags = QRhiTexture::Flags()) const;
+    bool isTextureFormatSupported(QRhiTexture::Format format, QRhiTexture::Flags flags = {}) const;
     bool isFeatureSupported(QRhi::Feature feature) const;
     int resourceLimit(ResourceLimit limit) const;
 
@@ -1518,6 +1641,9 @@ public:
 
     bool isDeviceLost() const;
 
+    QByteArray pipelineCacheData();
+    void setPipelineCacheData(const QByteArray &data);
+
 protected:
     QRhi();
 
