@@ -48,6 +48,7 @@
 #include <QtCore/qmetaobject.h>
 #include <QtCore/qsharedpointer.h>
 #include <QtCore/qtemporarydir.h>
+#include <QtCore/qthread.h>
 
 #include <string.h>
 
@@ -57,6 +58,7 @@
 
 QT_BEGIN_NAMESPACE
 
+class qfloat16;
 class QRegularExpression;
 
 #define QVERIFY(statement) \
@@ -67,17 +69,17 @@ do {\
 
 #define QFAIL(message) \
 do {\
-    QTest::qFail(message, __FILE__, __LINE__);\
+    QTest::qFail(static_cast<const char *>(message), __FILE__, __LINE__);\
     return;\
 } while (false)
 
 #define QVERIFY2(statement, description) \
 do {\
     if (statement) {\
-        if (!QTest::qVerify(true, #statement, (description), __FILE__, __LINE__))\
+        if (!QTest::qVerify(true, #statement, static_cast<const char *>(description), __FILE__, __LINE__))\
             return;\
     } else {\
-        if (!QTest::qVerify(false, #statement, (description), __FILE__, __LINE__))\
+        if (!QTest::qVerify(false, #statement, static_cast<const char *>(description), __FILE__, __LINE__))\
             return;\
     }\
 } while (false)
@@ -123,11 +125,11 @@ do {\
  * in their code.
  */
 #  define QVERIFY_EXCEPTION_THROWN(expression, exceptiontype) \
-    Q_STATIC_ASSERT_X(false, "Support of exceptions is disabled")
+    static_assert(false, "Support of exceptions is disabled")
 
 #endif // !QT_NO_EXCEPTIONS
 
-
+// NB: not do {...} while (0) wrapped, as qt_test_i is accessed after it
 #define QTRY_LOOP_IMPL(expr, timeoutValue, step) \
     if (!(expr)) { \
         QTest::qWait(0); \
@@ -150,10 +152,10 @@ do {\
 // Ideally we'd use qWaitFor instead of QTRY_LOOP_IMPL, but due
 // to a compiler bug on MSVC < 2017 we can't (see QTBUG-59096)
 #define QTRY_IMPL(expr, timeout)\
-    const int qt_test_step = 50; \
+    const int qt_test_step = timeout < 350 ? timeout / 7 + 1 : 50; \
     const int qt_test_timeoutValue = timeout; \
-    QTRY_LOOP_IMPL((expr), qt_test_timeoutValue, qt_test_step); \
-    QTRY_TIMEOUT_DEBUG_IMPL((expr), qt_test_timeoutValue, qt_test_step)\
+    { QTRY_LOOP_IMPL(QTest::currentTestFailed() || (expr), qt_test_timeoutValue, qt_test_step); } \
+    QTRY_TIMEOUT_DEBUG_IMPL(QTest::currentTestFailed() || (expr), qt_test_timeoutValue, qt_test_step)
 
 // Will try to wait for the expression to become true while allowing event processing
 #define QTRY_VERIFY_WITH_TIMEOUT(expr, timeout) \
@@ -184,23 +186,15 @@ do { \
 
 #define QSKIP_INTERNAL(statement) \
 do {\
-    QTest::qSkip(statement, __FILE__, __LINE__);\
+    QTest::qSkip(static_cast<const char *>(statement), __FILE__, __LINE__);\
     return;\
 } while (false)
 
-#ifdef Q_COMPILER_VARIADIC_MACROS
-
 #define QSKIP(statement, ...) QSKIP_INTERNAL(statement)
 
-#else
-
-#define QSKIP(statement) QSKIP_INTERNAL(statement)
-
-#endif
-
 #define QEXPECT_FAIL(dataIndex, comment, mode)\
 do {\
-    if (!QTest::qExpectFail(dataIndex, comment, QTest::mode, __FILE__, __LINE__))\
+    if (!QTest::qExpectFail(dataIndex, static_cast<const char *>(comment), QTest::mode, __FILE__, __LINE__))\
         return;\
 } while (false)
 
@@ -217,11 +211,16 @@ do {\
 } while (false)
 
 #define QWARN(msg)\
-    QTest::qWarn(msg, __FILE__, __LINE__)
+    QTest::qWarn(static_cast<const char *>(msg), __FILE__, __LINE__)
 
 #ifdef QT_TESTCASE_BUILDDIR
+
+#ifndef QT_TESTCASE_SOURCEDIR
+#define QT_TESTCASE_SOURCEDIR nullptr
+#endif
+
 # define QFINDTESTDATA(basepath)\
-    QTest::qFindTestData(basepath, __FILE__, __LINE__, QT_TESTCASE_BUILDDIR)
+    QTest::qFindTestData(basepath, __FILE__, __LINE__, QT_TESTCASE_BUILDDIR, QT_TESTCASE_SOURCEDIR)
 #else
 # define QFINDTESTDATA(basepath)\
     QTest::qFindTestData(basepath, __FILE__, __LINE__)
@@ -247,8 +246,14 @@ namespace QTest
         return qstrdup(me.valueToKey(int(e))); // int cast is necessary to support enum classes
     }
 
+    template <typename T>
+    inline typename std::enable_if<!QtPrivate::IsQEnumHelper<T>::Value && std::is_enum_v<T>, char*>::type toString(const T &e)
+    {
+        return qstrdup(QByteArray::number(static_cast<std::underlying_type_t<T>>(e)).constData());
+    }
+
     template <typename T> // Fallback
-    inline typename std::enable_if<!QtPrivate::IsQEnumHelper<T>::Value, char*>::type toString(const T &)
+    inline typename std::enable_if<!QtPrivate::IsQEnumHelper<T>::Value && !std::is_enum_v<T>, char*>::type toString(const T &)
     {
         return nullptr;
     }
@@ -290,7 +295,8 @@ namespace QTest
     Q_TESTLIB_EXPORT char *toPrettyCString(const char *unicode, int length);
     Q_TESTLIB_EXPORT char *toPrettyUnicode(QStringView string);
     Q_TESTLIB_EXPORT char *toString(const char *);
-    Q_TESTLIB_EXPORT char *toString(const void *);
+    Q_TESTLIB_EXPORT char *toString(const volatile void *);
+    Q_TESTLIB_EXPORT char *toString(const void *); // ### FIXME: Qt 7: Remove
 
     Q_TESTLIB_EXPORT void qInit(QObject *testObject, int argc = 0, char **argv = nullptr);
     Q_TESTLIB_EXPORT int qRun();
@@ -316,8 +322,8 @@ namespace QTest
 #if QT_CONFIG(temporaryfile)
     Q_TESTLIB_EXPORT QSharedPointer<QTemporaryDir> qExtractTestData(const QString &dirName);
 #endif
-    Q_TESTLIB_EXPORT QString qFindTestData(const char* basepath, const char* file = nullptr, int line = 0, const char* builddir = nullptr);
-    Q_TESTLIB_EXPORT QString qFindTestData(const QString& basepath, const char* file = nullptr, int line = 0, const char* builddir = nullptr);
+    Q_TESTLIB_EXPORT QString qFindTestData(const char* basepath, const char* file = nullptr, int line = 0, const char* builddir = nullptr, const char* sourcedir = nullptr);
+    Q_TESTLIB_EXPORT QString qFindTestData(const QString& basepath, const char* file = nullptr, int line = 0, const char* builddir = nullptr, const char *sourcedir = nullptr);
 
     Q_TESTLIB_EXPORT void *qData(const char *tagName, int typeId);
     Q_TESTLIB_EXPORT void *qGlobalData(const char *tagName, int typeId);
@@ -337,14 +343,13 @@ namespace QTest
                                          char *val1, char *val2,
                                          const char *actual, const char *expected,
                                          const char *file, int line);
-    Q_TESTLIB_EXPORT void qSleep(int ms);
     Q_TESTLIB_EXPORT void addColumnInternal(int id, const char *name);
 
     template <typename T>
     inline void addColumn(const char *name, T * = nullptr)
     {
-        typedef std::is_same<T, const char*> QIsSameTConstChar;
-        Q_STATIC_ASSERT_X(!QIsSameTConstChar::value, "const char* is not allowed as a test data format.");
+        using QIsSameTConstChar = std::is_same<T, const char*>;
+        static_assert(!QIsSameTConstChar::value, "const char* is not allowed as a test data format.");
         addColumnInternal(qMetaTypeId<T>(), name);
     }
     Q_TESTLIB_EXPORT QTestData &newRow(const char *dataTag);
@@ -361,12 +366,54 @@ namespace QTest
     }
 #endif
 
+    Q_TESTLIB_EXPORT bool qCompare(qfloat16 const &t1, qfloat16 const &t2,
+                    const char *actual, const char *expected, const char *file, int line);
+
     Q_TESTLIB_EXPORT bool qCompare(float const &t1, float const &t2,
                     const char *actual, const char *expected, const char *file, int line);
 
     Q_TESTLIB_EXPORT bool qCompare(double const &t1, double const &t2,
                     const char *actual, const char *expected, const char *file, int line);
 
+    Q_TESTLIB_EXPORT bool qCompare(int t1, int t2, const char *actual, const char *expected,
+                                   const char *file, int line);
+
+#if QT_POINTER_SIZE == 8
+    Q_TESTLIB_EXPORT bool qCompare(qsizetype t1, qsizetype t2, const char *actual, const char *expected,
+                                   const char *file, int line);
+#endif
+
+    Q_TESTLIB_EXPORT bool qCompare(unsigned t1, unsigned t2, const char *actual, const char *expected,
+                                   const char *file, int line);
+
+    Q_TESTLIB_EXPORT bool qCompare(QStringView t1, QStringView t2,
+                                   const char *actual, const char *expected,
+                                   const char *file, int line);
+    Q_TESTLIB_EXPORT bool qCompare(QStringView t1, const QLatin1String &t2,
+                                   const char *actual, const char *expected,
+                                   const char *file, int line);
+    Q_TESTLIB_EXPORT bool qCompare(const QLatin1String &t1, QStringView t2,
+                                   const char *actual, const char *expected,
+                                   const char *file, int line);
+    inline bool qCompare(const QString &t1, const QString &t2,
+                         const char *actual, const char *expected,
+                         const char *file, int line)
+    {
+        return qCompare(QStringView(t1), QStringView(t2), actual, expected, file, line);
+    }
+    inline bool qCompare(const QString &t1, const QLatin1String &t2,
+                         const char *actual, const char *expected,
+                         const char *file, int line)
+    {
+        return qCompare(QStringView(t1), t2, actual, expected, file, line);
+    }
+    inline bool qCompare(const QLatin1String &t1, const QString &t2,
+                         const char *actual, const char *expected,
+                         const char *file, int line)
+    {
+        return qCompare(t1, QStringView(t2), actual, expected, file, line);
+    }
+
     inline bool compare_ptr_helper(const volatile void *t1, const volatile void *t2, const char *actual,
                                    const char *expected, const char *file, int line)
     {
@@ -405,6 +452,7 @@ namespace QTest
 
     QTEST_COMPARE_DECL(float)
     QTEST_COMPARE_DECL(double)
+    QTEST_COMPARE_DECL(qfloat16)
     QTEST_COMPARE_DECL(char)
     QTEST_COMPARE_DECL(signed char)
     QTEST_COMPARE_DECL(unsigned char)
