@@ -49,8 +49,6 @@
 #endif
 
 #include "qevent.h"
-#include "qdesktopwidget.h"
-#include <private/qdesktopwidget_p.h>
 #include "qapplication.h"
 #include "qlayout.h"
 #if QT_CONFIG(sizegrip)
@@ -148,7 +146,7 @@ bool QDialogPrivate::canBeNativeDialog() const
 /*!
     \internal
 
-    Properly hides dialog and sets the \p resultCode
+    Properly hides dialog and sets the \a resultCode.
  */
 void QDialogPrivate::hide(int resultCode)
 {
@@ -164,8 +162,8 @@ void QDialogPrivate::hide(int resultCode)
 /*!
     \internal
 
-    Emits finished() signal with \p resultCode. If the \p dialogCode
-    is equal to 0 emits rejected(), if the \p dialogCode is equal to
+    Emits finished() signal with \a resultCode. If the \a dialogCode
+    is equal to 0 emits rejected(), if the \a dialogCode is equal to
     1 emits accepted().
  */
 void QDialogPrivate::finalize(int resultCode, int dialogCode)
@@ -187,7 +185,7 @@ QWindow *QDialogPrivate::transientParentWindow() const
         return parent->windowHandle();
     else if (q->windowHandle())
         return q->windowHandle()->transientParent();
-    return 0;
+    return nullptr;
 }
 
 bool QDialogPrivate::setNativeDialogVisible(bool visible)
@@ -214,7 +212,7 @@ QVariant QDialogPrivate::styleHint(QPlatformDialogHelper::StyleHint hint) const
 void QDialogPrivate::deletePlatformHelper()
 {
     delete m_platformHelper;
-    m_platformHelper = 0;
+    m_platformHelper = nullptr;
     m_platformHelperCreated = false;
     nativeDialogInUse = false;
 }
@@ -246,6 +244,13 @@ void QDialogPrivate::deletePlatformHelper()
     window-system properties for the widget (in particular it will
     reset the Qt::Dialog flag).
 
+    \note The parent relationship of the dialog does \e{not} imply
+    that the dialog will always be stacked on top of the parent
+    window. To ensure that the dialog is always on top, make the
+    dialog modal. This also applies for child windows of the dialog
+    itself. To ensure that child windows of the dialog stay on top
+    of the dialog, make the child windows modal as well.
+
     \section1 Modal Dialogs
 
     A \b{modal} dialog is a dialog that blocks input to other
@@ -263,12 +268,11 @@ void QDialogPrivate::deletePlatformHelper()
 
     The most common way to display a modal dialog is to call its
     exec() function. When the user closes the dialog, exec() will
-    provide a useful \l{#return}{return value}. Typically,
-    to get the dialog to close and return the appropriate value, we
-    connect a default button, e.g. \uicontrol OK, to the accept() slot and a
-    \uicontrol Cancel button to the reject() slot.
-    Alternatively you can call the done() slot with \c Accepted or
-    \c Rejected.
+    provide a useful \l{#return}{return value}. To close the dialog
+    and return the appropriate value, you must connect a default button,
+    e.g. an \uicontrol OK button to the accept() slot and a
+    \uicontrol Cancel button to the reject() slot. Alternatively, you
+    can call the done() slot with \c Accepted or \c Rejected.
 
     An alternative is to call setModal(true) or setWindowModality(),
     then show(). Unlike exec(), show() returns control to the caller
@@ -276,7 +280,7 @@ void QDialogPrivate::deletePlatformHelper()
     progress dialogs, where the user must have the ability to interact
     with the dialog, e.g.  to cancel a long running operation. If you
     use show() and setModal(true) together to perform a long operation,
-    you must call QApplication::processEvents() periodically during
+    you must call QCoreApplication::processEvents() periodically during
     processing to enable the user to interact with the dialog. (See
     QProgressDialog.)
 
@@ -422,7 +426,7 @@ QDialog::~QDialog()
 /*!
   \internal
   This function is called by the push button \a pushButton when it
-  becomes the default button. If \a pushButton is 0, the dialogs
+  becomes the default button. If \a pushButton is \nullptr, the dialogs
   default default button becomes the default button. This is what a
   push button calls when it loses focus.
 */
@@ -454,7 +458,7 @@ void QDialogPrivate::setDefault(QPushButton *pushButton)
 */
 void QDialogPrivate::setMainDefault(QPushButton *pushButton)
 {
-    mainDef = 0;
+    mainDef = nullptr;
     setDefault(pushButton);
 }
 
@@ -596,7 +600,7 @@ int QDialog::exec()
     }
     if (guard.isNull())
         return QDialog::Rejected;
-    d->eventLoop = 0;
+    d->eventLoop = nullptr;
 
     setAttribute(Qt::WA_ShowModal, wasShowModal);
 
@@ -673,19 +677,19 @@ void QDialog::contextMenuEvent(QContextMenuEvent *e)
 #else
     QWidget *w = childAt(e->pos());
     if (!w) {
-        w = rect().contains(e->pos()) ? this : 0;
+        w = rect().contains(e->pos()) ? this : nullptr;
         if (!w)
             return;
     }
     while (w && w->whatsThis().size() == 0 && !w->testAttribute(Qt::WA_CustomWhatsThis))
-        w = w->isWindow() ? 0 : w->parentWidget();
+        w = w->isWindow() ? nullptr : w->parentWidget();
     if (w) {
         QPointer<QMenu> p = new QMenu(this);
         QAction *wt = p.data()->addAction(tr("What's This?"));
         if (p.data()->exec(e->globalPos()) == wt) {
             QHelpEvent e(QEvent::WhatsThis, w->rect().center(),
                          w->mapToGlobal(w->rect().center()));
-            QApplication::sendEvent(w, &e);
+            QCoreApplication::sendEvent(w, &e);
         }
         delete p.data();
     }
@@ -760,12 +764,25 @@ void QDialog::setVisible(bool visible)
     if (!testAttribute(Qt::WA_DontShowOnScreen) && d->canBeNativeDialog() && d->setNativeDialogVisible(visible))
         return;
 
+    // We should not block windows by the invisible modal dialog
+    // if a platform-specific dialog is implemented as an in-process
+    // Qt window, because in this case it will also be blocked.
+    const bool dontBlockWindows = testAttribute(Qt::WA_DontShowOnScreen)
+            && d->styleHint(QPlatformDialogHelper::DialogIsQtWindow).toBool();
+    Qt::WindowModality oldModality;
+    bool wasModalitySet;
+
+    if (dontBlockWindows) {
+        oldModality = windowModality();
+        wasModalitySet = testAttribute(Qt::WA_SetWindowModality);
+        setWindowModality(Qt::NonModal);
+    }
+
     if (visible) {
         if (testAttribute(Qt::WA_WState_ExplicitShowHide) && !testAttribute(Qt::WA_WState_Hidden))
             return;
 
         QWidget::setVisible(visible);
-        showExtension(d->doShowExtension);
         QWidget *fw = window()->focusWidget();
         if (!fw)
             fw = this;
@@ -801,7 +818,7 @@ void QDialog::setVisible(bool visible)
 #endif
         if (fw && !fw->hasFocus()) {
             QFocusEvent e(QEvent::FocusIn, Qt::TabFocusReason);
-            QApplication::sendEvent(fw, &e);
+            QCoreApplication::sendEvent(fw, &e);
         }
 
 #ifndef QT_NO_ACCESSIBILITY
@@ -826,6 +843,11 @@ void QDialog::setVisible(bool visible)
             d->eventLoop->exit();
     }
 
+    if (dontBlockWindows) {
+        setWindowModality(oldModality);
+        setAttribute(Qt::WA_SetWindowModality, wasModalitySet);
+    }
+
 #if QT_CONFIG(pushbutton)
     const QPlatformTheme *theme = QGuiApplicationPrivate::platformTheme();
     if (d->mainDef && isActiveWindow()
@@ -849,23 +871,31 @@ void QDialog::showEvent(QShowEvent *event)
 /*! \internal */
 void QDialog::adjustPosition(QWidget* w)
 {
+    Q_D(QDialog);
 
     if (const QPlatformTheme *theme = QGuiApplicationPrivate::platformTheme())
         if (theme->themeHint(QPlatformTheme::WindowAutoPlacement).toBool())
             return;
     QPoint p(0, 0);
-    int extraw = 0, extrah = 0, scrn = 0;
-    if (w)
-        w = w->window();
-    QRect desk;
+    int extraw = 0, extrah = 0;
+    const QWindow *parentWindow = nullptr;
     if (w) {
-        scrn = QDesktopWidgetPrivate::screenNumber(w);
-    } else if (QDesktopWidgetPrivate::isVirtualDesktop()) {
-        scrn = QDesktopWidgetPrivate::screenNumber(QCursor::pos());
+        w = w->window();
     } else {
-        scrn = QDesktopWidgetPrivate::screenNumber(this);
+        parentWindow = d->transientParentWindow();
     }
-    desk = QDesktopWidgetPrivate::availableGeometry(scrn);
+    QRect desk;
+    QScreen *scrn = nullptr;
+    if (w)
+        scrn = w->screen();
+    else if (parentWindow)
+        scrn = parentWindow->screen();
+    else if (QGuiApplication::primaryScreen()->virtualSiblings().size() > 1)
+        scrn = QGuiApplication::screenAt(QCursor::pos());
+    else
+        scrn = screen();
+    if (scrn)
+        desk = scrn->availableGeometry();
 
     QWidgetList list = QApplication::topLevelWidgets();
     for (int i = 0; (extraw == 0 || extrah == 0) && i < list.size(); ++i) {
@@ -890,12 +920,17 @@ void QDialog::adjustPosition(QWidget* w)
     if (w) {
         // Use pos() if the widget is embedded into a native window
         QPoint pp;
-        if (w->windowHandle() && w->windowHandle()->property("_q_embedded_native_parent_handle").value<WId>())
+        if (w->windowHandle() && qvariant_cast<WId>(w->windowHandle()->property("_q_embedded_native_parent_handle")))
             pp = w->pos();
         else
             pp = w->mapToGlobal(QPoint(0,0));
         p = QPoint(pp.x() + w->width()/2,
                     pp.y() + w->height()/ 2);
+    } else if (parentWindow) {
+        // QTBUG-63406: Widget-based dialog in QML, which has no Widget parent
+        // but a transient parent window.
+        QPoint pp = parentWindow->mapToGlobal(QPoint(0, 0));
+        p = QPoint(pp.x() + parentWindow->width() / 2, pp.y() + parentWindow->height() / 2);
     } else {
         // p = middle of the desktop
         p = QPoint(desk.x() + desk.width()/2, desk.y() + desk.height()/2);
@@ -919,160 +954,14 @@ void QDialog::adjustPosition(QWidget* w)
     // QTBUG-52735: Manually set the correct target screen since scaling in a
     // subsequent call to QWindow::resize() may otherwise use the wrong factor
     // if the screen changed notification is still in an event queue.
-    if (scrn >= 0) {
+    if (scrn) {
         if (QWindow *window = windowHandle())
-            window->setScreen(QGuiApplication::screens().at(scrn));
+            window->setScreen(scrn);
     }
 
     move(p);
 }
 
-/*!
-    \obsolete
-
-    If \a orientation is Qt::Horizontal, the extension will be displayed
-    to the right of the dialog's main area. If \a orientation is
-    Qt::Vertical, the extension will be displayed below the dialog's main
-    area.
-
-    Instead of using this functionality, we recommend that you simply call
-    show() or hide() on the part of the dialog that you want to use as an
-    extension. See the \l{Extension Example} for details.
-
-    \sa setExtension()
-*/
-void QDialog::setOrientation(Qt::Orientation orientation)
-{
-    Q_D(QDialog);
-    d->orientation = orientation;
-}
-
-/*!
-    \obsolete
-
-    Returns the dialog's extension orientation.
-
-    Instead of using this functionality, we recommend that you simply call
-    show() or hide() on the part of the dialog that you want to use as an
-    extension. See the \l{Extension Example} for details.
-
-    \sa extension()
-*/
-Qt::Orientation QDialog::orientation() const
-{
-    Q_D(const QDialog);
-    return d->orientation;
-}
-
-/*!
-    \obsolete
-
-    Sets the widget, \a extension, to be the dialog's extension,
-    deleting any previous extension. The dialog takes ownership of the
-    extension. Note that if 0 is passed any existing extension will be
-    deleted. This function must only be called while the dialog is hidden.
-
-    Instead of using this functionality, we recommend that you simply call
-    show() or hide() on the part of the dialog that you want to use as an
-    extension. See the \l{Extension Example} for details.
-
-    \sa showExtension(), setOrientation()
-*/
-void QDialog::setExtension(QWidget* extension)
-{
-    Q_D(QDialog);
-    delete d->extension;
-    d->extension = extension;
-
-    if (!extension)
-        return;
-
-    if (extension->parentWidget() != this)
-        extension->setParent(this);
-    extension->hide();
-}
-
-/*!
-    \obsolete
-
-    Returns the dialog's extension or 0 if no extension has been
-    defined.
-
-    Instead of using this functionality, we recommend that you simply call
-    show() or hide() on the part of the dialog that you want to use as an
-    extension. See the \l{Extension Example} for details.
-
-    \sa showExtension(), setOrientation()
-*/
-QWidget* QDialog::extension() const
-{
-    Q_D(const QDialog);
-    return d->extension;
-}
-
-
-/*!
-    \obsolete
-
-    If \a showIt is true, the dialog's extension is shown; otherwise the
-    extension is hidden.
-
-    Instead of using this functionality, we recommend that you simply call
-    show() or hide() on the part of the dialog that you want to use as an
-    extension. See the \l{Extension Example} for details.
-
-    \sa show(), setExtension(), setOrientation()
-*/
-void QDialog::showExtension(bool showIt)
-{
-    Q_D(QDialog);
-    d->doShowExtension = showIt;
-    if (!d->extension)
-        return;
-    if (!testAttribute(Qt::WA_WState_Visible))
-        return;
-    if (d->extension->isVisible() == showIt)
-        return;
-
-    if (showIt) {
-        d->size = size();
-        d->min = minimumSize();
-        d->max = maximumSize();
-        if (layout())
-            layout()->setEnabled(false);
-        QSize s(d->extension->sizeHint()
-                 .expandedTo(d->extension->minimumSize())
-                 .boundedTo(d->extension->maximumSize()));
-        if (d->orientation == Qt::Horizontal) {
-            int h = qMax(height(), s.height());
-            d->extension->setGeometry(width(), 0, s.width(), h);
-            setFixedSize(width() + s.width(), h);
-        } else {
-            int w = qMax(width(), s.width());
-            d->extension->setGeometry(0, height(), w, s.height());
-            setFixedSize(w, height() + s.height());
-        }
-        d->extension->show();
-#if QT_CONFIG(sizegrip)
-        const bool sizeGripEnabled = isSizeGripEnabled();
-        setSizeGripEnabled(false);
-        d->sizeGripEnabled = sizeGripEnabled;
-#endif
-    } else {
-        d->extension->hide();
-        // workaround for CDE window manager that won't shrink with (-1,-1)
-        setMinimumSize(d->min.expandedTo(QSize(1, 1)));
-        setMaximumSize(d->max);
-        resize(d->size);
-        if (layout())
-            layout()->setEnabled(true);
-#if QT_CONFIG(sizegrip)
-        setSizeGripEnabled(d->sizeGripEnabled);
-#endif
-    }
-}
-
-
 /*! \reimp */
 QSize QDialog::sizeHint() const
 {
@@ -1160,7 +1049,7 @@ void QDialog::setSizeGripEnabled(bool enabled)
             d->resizer->show();
         } else {
             delete d->resizer;
-            d->resizer = 0;
+            d->resizer = nullptr;
         }
     }
 #endif // QT_CONFIG(sizegrip)
