@@ -27,7 +27,8 @@
 **
 ****************************************************************************/
 
-#include <QtTest/QtTest>
+#include <QTest>
+#include <QLibraryInfo>
 #include <QtCore/QString>
 #include <QtCore/QCoreApplication>
 #include <QtCore/QByteArray>
@@ -103,7 +104,7 @@ private:
 
 void tst_rcc::initTestCase()
 {
-    m_rcc = QLibraryInfo::path(QLibraryInfo::BinariesPath) + QLatin1String("/rcc");
+    m_rcc = QLibraryInfo::path(QLibraryInfo::LibraryExecutablesPath) + QLatin1String("/rcc");
 
     m_dataPath = QFINDTESTDATA("data");
     QVERIFY(!m_dataPath.isEmpty());
