@@ -27,8 +27,8 @@
 ****************************************************************************/
 
 
-#include <QtTest/QtTest>
-
+#include <QTest>
+#include <QSignalSpy>
 
 #include "qpushbutton.h"
 #include <qapplication.h>
@@ -53,7 +53,9 @@ private slots:
     void getSetCheck();
     void autoRepeat();
     void pressed();
+#if QT_CONFIG(shortcut)
     void setAccel();
+#endif
     void isCheckable();
     void setDown();
     void popupCrash();
@@ -61,12 +63,17 @@ private slots:
     void animateClick();
     void toggle();
     void clicked();
+    void touchTap();
     void toggled();
     void defaultAndAutoDefault();
     void sizeHint_data();
     void sizeHint();
+#if QT_CONFIG(shortcut)
     void taskQTBUG_20191_shortcutWithKeypadModifer();
+#endif
     void emitReleasedAfterChange();
+    void hitButton();
+    void iconOnlyStyleSheet();
 
 protected slots:
     void resetCounters();
@@ -83,6 +90,7 @@ private:
     uint release_count;
 
     QPushButton *testWidget;
+    QPointingDevice *m_touchScreen = QTest::createTouchDevice();
 };
 
 // Testing get/set functions
@@ -125,8 +133,10 @@ void tst_QPushButton::init()
     testWidget->setDown( false );
     testWidget->setText("Test");
     testWidget->setEnabled( true );
+#if QT_CONFIG(shortcut)
     QKeySequence seq;
     testWidget->setShortcut( seq );
+#endif
 
     resetCounters();
 }
@@ -183,9 +193,7 @@ void tst_QPushButton::autoRepeat()
     testWidget->setAutoRepeat( false );
     QTest::keyPress( testWidget, Qt::Key_Space );
 
-    QTest::qWait( 300 );
-
-    QVERIFY( testWidget->isDown() );
+    QTRY_VERIFY( testWidget->isDown() );
     QVERIFY( toggle_count == 0 );
     QVERIFY( press_count == 1 );
     QVERIFY( release_count == 0 );
@@ -200,13 +208,12 @@ void tst_QPushButton::autoRepeat()
     testWidget->setDown( false );
     testWidget->setAutoRepeat( true );
     QTest::keyPress( testWidget, Qt::Key_Space );
-    QTest::qWait(900);
+    QTRY_VERIFY(press_count > 3);
     QVERIFY( testWidget->isDown() );
     QVERIFY( toggle_count == 0 );
     QTest::keyRelease( testWidget, Qt::Key_Space );
     QCOMPARE(press_count, release_count);
     QCOMPARE(release_count, click_count);
-    QVERIFY(press_count > 1);
 
     // #### shouldn't I check here to see if multiple signals have been fired???
 
@@ -241,23 +248,19 @@ void tst_QPushButton::autoRepeat()
 void tst_QPushButton::pressed()
 {
     QTest::keyPress( testWidget, ' ' );
-//    QTest::qWait( 300 );
     QCOMPARE( press_count, (uint)1 );
     QCOMPARE( release_count, (uint)0 );
 
     QTest::keyRelease( testWidget, ' ' );
-//    QTest::qWait( 300 );
     QCOMPARE( press_count, (uint)1 );
     QCOMPARE( release_count, (uint)1 );
 
     QTest::keyPress( testWidget,Qt::Key_Enter );
-//    QTest::qWait( 300 );
     QCOMPARE( press_count, (uint)1 );
     QCOMPARE( release_count, (uint)1 );
 
     testWidget->setAutoDefault(true);
     QTest::keyPress( testWidget,Qt::Key_Enter );
-//    QTest::qWait( 300 );
     QCOMPARE( press_count, (uint)2 );
     QCOMPARE( release_count, (uint)2 );
     testWidget->setAutoDefault(false);
@@ -325,6 +328,8 @@ void tst_QPushButton::toggled()
     QVERIFY( click_count == 1 );
 }
 
+#if QT_CONFIG(shortcut)
+
 /*
     If we press an accelerator key we ONLY get a pressed signal and
     NOT a released or clicked signal.
@@ -332,22 +337,19 @@ void tst_QPushButton::toggled()
 
 void tst_QPushButton::setAccel()
 {
+    if (QGuiApplication::platformName().startsWith(QLatin1String("wayland"), Qt::CaseInsensitive))
+        QSKIP("Wayland: This fails. Figure out why.");
+
     testWidget->setText("&AccelTest");
-    QKeySequence seq( Qt::ALT + Qt::Key_A );
+    QKeySequence seq( Qt::ALT | Qt::Key_A );
     testWidget->setShortcut( seq );
 
     // The shortcut will not be activated unless the button is in a active
     // window and has focus
     QApplication::setActiveWindow(testWidget);
     testWidget->setFocus();
-    for (int i = 0; !testWidget->isActiveWindow() && i < 1000; ++i) {
-        testWidget->activateWindow();
-        QApplication::instance()->processEvents();
-        QTest::qWait(100);
-    }
-    QVERIFY(testWidget->isActiveWindow());
+    QVERIFY(QTest::qWaitForWindowActive(testWidget));
     QTest::keyClick( testWidget, 'A', Qt::AltModifier );
-    QTest::qWait( 50 );
     QTRY_VERIFY( click_count == 1 );
     QVERIFY( press_count == 1 );
     QVERIFY( release_count == 1 );
@@ -359,6 +361,8 @@ void tst_QPushButton::setAccel()
     QTRY_VERIFY( !testWidget->isDown() );
 }
 
+#endif // QT_CONFIG(shortcut)
+
 void tst_QPushButton::animateClick()
 {
     QVERIFY( !testWidget->isDown() );
@@ -392,6 +396,29 @@ void tst_QPushButton::clicked()
     QCOMPARE( release_count, (uint)10 );
 }
 
+void tst_QPushButton::touchTap()
+{
+    QTest::touchEvent(testWidget, m_touchScreen).press(0, QPoint(10, 10));
+    QVERIFY( press_count == 1 );
+    QVERIFY( release_count == 0 );
+    QTest::touchEvent(testWidget, m_touchScreen).release(0, QPoint(10, 10));
+    QCOMPARE( press_count, (uint)1 );
+    QCOMPARE( release_count, (uint)1 );
+    QCOMPARE( click_count, (uint)1 );
+
+    press_count = 0;
+    release_count = 0;
+    click_count = 0;
+    testWidget->setDown(false);
+    for (uint i = 0; i < 10; i++) {
+        QTest::touchEvent(testWidget, m_touchScreen).press(0, QPoint(10, 10));
+        QTest::touchEvent(testWidget, m_touchScreen).release(0, QPoint(10, 10));
+    }
+    QCOMPARE( press_count, (uint)10 );
+    QCOMPARE( release_count, (uint)10 );
+    QCOMPARE( click_count, (uint)10 );
+}
+
 QPushButton *pb = 0;
 void tst_QPushButton::helperSlotDelete()
 {
@@ -515,7 +542,7 @@ void tst_QPushButton::sizeHint_data()
     QTest::newRow("windows") << QString::fromLatin1("windows");
 #endif
 #if defined(Q_OS_MAC) && !defined(QT_NO_STYLE_MAC)
-    QTest::newRow("macintosh") << QString::fromLatin1("macintosh");
+    QTest::newRow("macos") << QString::fromLatin1("macos");
 #endif
 #if !defined(QT_NO_STYLE_FUSION)
     QTest::newRow("fusion") << QString::fromLatin1("fusion");
@@ -576,13 +603,13 @@ void tst_QPushButton::sizeHint()
         dialog->showNormal();
         tabWidget->setCurrentWidget(tab2);
         tabWidget->setCurrentWidget(tab1);
-        QTest::qWait(100);
-        QApplication::processEvents();
 
-        QCOMPARE(button1_2->size(), button2_2->size());
+        QTRY_COMPARE(button1_2->size(), button2_2->size());
     }
 }
 
+#if QT_CONFIG(shortcut)
+
 void tst_QPushButton::taskQTBUG_20191_shortcutWithKeypadModifer()
 {
     // setup a dialog with two buttons
@@ -609,7 +636,7 @@ void tst_QPushButton::taskQTBUG_20191_shortcutWithKeypadModifer()
     // add shortcut 'keypad 5' to button2
     spy1.clear();
     QSignalSpy spy2(button2, SIGNAL(clicked()));
-    button2->setShortcut(Qt::Key_5 + Qt::KeypadModifier);
+    button2->setShortcut(Qt::Key_5 | Qt::KeypadModifier);
     QTest::keyClick(&dialog, Qt::Key_5);
     QTest::qWait(300);
     QTest::keyClick(&dialog, Qt::Key_5, Qt::KeypadModifier);
@@ -629,6 +656,8 @@ void tst_QPushButton::taskQTBUG_20191_shortcutWithKeypadModifer()
     QCOMPARE(spy2.count(), 1);
 }
 
+#endif // QT_CONFIG(shortcut)
+
 void tst_QPushButton::emitReleasedAfterChange()
 {
     QPushButton *button1 = new QPushButton("A");
@@ -660,5 +689,68 @@ void tst_QPushButton::emitReleasedAfterChange()
     QCOMPARE(spy.count(), 1);
 }
 
+/*
+    Test that QPushButton::hitButton returns true for points that
+    are certainly inside the bevel, also when a style sheet is set.
+*/
+void tst_QPushButton::hitButton()
+{
+    class PushButton : public QPushButton
+    {
+    public:
+        PushButton(const QString &text = {})
+        : QPushButton(text)
+        {}
+
+        bool hitButton(const QPoint &point) const override
+        {
+            return QPushButton::hitButton(point);
+        }
+    };
+
+    QDialog dialog;
+    QVBoxLayout *layout = new QVBoxLayout;
+    PushButton *button1 = new PushButton("Ok");
+    PushButton *button2 = new PushButton("Cancel");
+    button2->setStyleSheet("QPushButton {"
+        "padding: 5px;"
+        "margin: 5px;"
+        "border-radius: 4px;"
+        "border: 1px solid black; }"
+    );
+
+    layout->addWidget(button1);
+    layout->addWidget(button2);
+
+    dialog.setLayout(layout);
+    dialog.show();
+    QVERIFY(QTest::qWaitForWindowExposed(&dialog));
+
+    const QPoint button1Center = button1->rect().center();
+    QVERIFY(button1->hitButton(button1Center));
+
+    const QPoint button2Center = button2->rect().center();
+    QVERIFY(button2->hitButton(button2Center));
+    QVERIFY(button2->hitButton(QPoint(6, 6)));
+    QVERIFY(!button2->hitButton(QPoint(2, 2)));
+}
+
+/*
+    Test that a style sheet with only icon doesn't crash.
+    QTBUG-91735
+*/
+void tst_QPushButton::iconOnlyStyleSheet()
+{
+    QIcon icon(":/qt-project.org/styles/commonstyle/images/dvd-32.png");
+    QVERIFY(!icon.isNull());
+    QPushButton pb;
+    pb.setStyleSheet("QPushButton {"
+        "icon: url(:/qt-project.org/styles/commonstyle/images/dvd-32.png);"
+        "border: red;"
+    "}");
+    pb.show();
+    QVERIFY(QTest::qWaitForWindowExposed(&pb));
+}
+
 QTEST_MAIN(tst_QPushButton)
 #include "tst_qpushbutton.moc"
